#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.31 16:00:00                  #
# ================================================== #

import mimetypes
from typing import Optional, Dict, Any, List
from google import genai
from google.genai import types as gtypes
from PySide6.QtCore import QObject, Signal, QRunnable, Slot
import base64, datetime, os, requests, tempfile

from pygpt_net.core.events import KernelEvent
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Image:

    MODE_GENERATE = "generate"
    MODE_EDIT = "edit"

    def __init__(self, window=None):
        self.window = window
        self.worker = None

    def generate(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None,
            sync: bool = True
    ) -> bool:
        """
        Generate or edit image(s) using Google GenAI API (Developer API or Vertex AI).
        """
        # Music fast-path: delegate to Music flow if a music model is selected (e.g., Lyria).
        try:
            model_id = (context.model.id if context and context.model else "") or ""
            if self.window and hasattr(self.window.core.api.google, "music"):
                if self.window.core.api.google.music.is_music_model(model_id):
                    return self.window.core.api.google.music.generate(context=context, extra=extra, sync=sync)
        except Exception:
            pass

        extra = extra or {}
        ctx = context.ctx or CtxItem()
        model = context.model
        prompt = context.prompt
        num = int(extra.get("num", 1))
        inline = bool(extra.get("inline", False))
        extra_prompt = extra.get("extra_prompt", "")

        # decide sub-mode based on attachments
        sub_mode = self.MODE_GENERATE
        attachments = context.attachments

        # model used to improve the prompt (not image model)
        prompt_model = self.window.core.models.from_defaults()
        tmp = self.window.core.config.get('img_prompt_model')
        if self.window.core.models.has(tmp):
            prompt_model = self.window.core.models.get(tmp)

        worker = ImageWorker()
        worker.window = self.window
        worker.client = self.window.core.api.google.get_client()
        worker.ctx = ctx
        worker.mode = sub_mode
        worker.attachments = attachments or {}
        worker.model = model.id  # image model id
        worker.input_prompt = prompt
        worker.model_prompt = prompt_model  # LLM for prompt rewriting
        worker.system_prompt = self.window.core.prompt.get('img')
        worker.raw = self.window.core.config.get('img_raw')
        worker.num = num
        worker.inline = inline
        worker.extra_prompt = extra_prompt

        # remix: previous image reference (ID/URI/path) from extra
        worker.image_id = extra.get("image_id")

        if attachments and len(attachments) > 0:
            mid = str(model.id).lower()
            if "imagen" in mid:
                worker.mode = self.MODE_EDIT

        if self.window.core.config.has('img_resolution'):
            worker.resolution = self.window.core.config.get('img_resolution') or "1024x1024"

        self.worker = worker
        self.worker.signals.finished.connect(self.window.core.image.handle_finished)
        self.worker.signals.finished_inline.connect(self.window.core.image.handle_finished_inline)
        self.worker.signals.status.connect(self.window.core.image.handle_status)
        self.worker.signals.error.connect(self.window.core.image.handle_error)

        if sync or not self.window.controller.kernel.async_allowed(ctx):
            self.worker.run()
            return True

        self.window.dispatch(KernelEvent(KernelEvent.STATE_BUSY, {"id": "img"}))
        self.window.threadpool.start(self.worker)
        return True


class ImageSignals(QObject):
    finished = Signal(object, list, str)         # ctx, paths, prompt
    finished_inline = Signal(object, list, str)  # ctx, paths, prompt
    status = Signal(object)                      # message
    error = Signal(object)                       # exception


class ImageWorker(QRunnable):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.signals = ImageSignals()
        self.window = None
        self.client: Optional[genai.Client] = None
        self.ctx: Optional[CtxItem] = None

        # params
        self.mode = Image.MODE_GENERATE
        self.attachments: Dict[str, Any] = {}
        self.model = "imagen-4.0-generate-001"
        self.model_prompt = None
        self.input_prompt = ""
        self.system_prompt = ""
        self.inline = False
        self.extra_prompt: Optional[str] = None
        self.raw = False
        self.num = 1
        self.resolution = "1024x1024"  # used to derive aspect ratio or image_size
        self.image_id: Optional[str] = None  # remix/extend previous image

        # limits
        self.imagen_max_num = 4  # Imagen returns up to 4 images

        # fallbacks
        self.DEFAULT_GEMINI_IMAGE_MODEL = "gemini-2.5-flash-image"

        # Canonical 1K dimensions for Nano Banana Pro (Gemini 3 Pro Image Preview).
        # Used to infer 2K/4K by 2x/4x multiples and to normalize UI inputs.
        self._NB_PRO_1K = {
            "1024x1024",  # 1:1
            "848x1264",   # 2:3
            "1264x848",   # 3:2
            "896x1200",   # 3:4
            "1200x896",   # 4:3
            "928x1152",   # 4:5
            "1152x928",   # 5:4
            "768x1376",   # 9:16
            "1376x768",   # 16:9
            "1584x672",   # 21:9
        }

    @Slot()
    def run(self):
        try:
            # optional prompt enhancement
            if not self.raw and not self.inline:
                try:
                    self.signals.status.emit(trans('img.status.prompt.wait'))
                    bridge_context = BridgeContext(
                        prompt=self.input_prompt,
                        system_prompt=self.system_prompt,
                        model=self.model_prompt,
                        max_tokens=200,
                        temperature=1.0,
                    )
                    ev = KernelEvent(KernelEvent.CALL, {'context': bridge_context, 'extra': {}})
                    self.window.dispatch(ev)
                    resp = ev.data.get('response')
                    if resp:
                        self.input_prompt = resp
                except Exception as e:
                    self.signals.error.emit(e)
                    self.signals.status.emit(trans('img.status.prompt.error') + ": " + str(e))

            # Decide how to apply negative prompt: native param on Vertex Imagen 3.0 (-001) or inline fallback.
            use_param = (
                bool(self.extra_prompt and str(self.extra_prompt).strip())
                and self._using_vertex()
                and self._imagen_supports_negative_prompt(self.model)
            )
            if (self.extra_prompt and str(self.extra_prompt).strip()) and not use_param:
                try:
                    self.input_prompt = self._merge_negative_prompt(self.input_prompt or "", self.extra_prompt)
                except Exception:
                    pass

            paths: List[str] = []

            # Remix path: if image_id provided, prefer image-to-image remix using the given identifier.
            if self.image_id:
                self.signals.status.emit(trans('img.status.generating') + " (remix): " + (self.input_prompt or "") + "...")
                if self._using_vertex() and self._is_imagen_generate(self.model):
                    # Vertex / Imagen edit flow with a single base image (no explicit mask).
                    img_ref = self._imagen_image_from_identifier(self.image_id)
                    if not img_ref:
                        raise RuntimeError("Invalid image_id for remix. Provide a valid local path, Files API name, or gs:// URI.")

                    raw_ref = gtypes.RawReferenceImage(reference_id=0, reference_image=img_ref)
                    mask_ref = gtypes.MaskReferenceImage(
                        reference_id=1,
                        reference_image=None,
                        config=gtypes.MaskReferenceConfig(
                            mask_mode="MASK_MODE_BACKGROUND",
                            mask_dilation=0.0,
                        ),
                    )
                    # Prepare edit config with optional negative prompt when supported
                    cfg_kwargs = dict(
                        edit_mode="EDIT_MODE_DEFAULT",
                        number_of_images=min(self.num, self.imagen_max_num),
                        include_rai_reason=True,
                    )
                    if self.extra_prompt and self._imagen_supports_negative_prompt(self.model):
                        cfg_kwargs["negative_prompt"] = self.extra_prompt
                    try:
                        cfg = gtypes.EditImageConfig(**cfg_kwargs)
                    except Exception:
                        # Fallback without negative_prompt if SDK doesn't recognize it
                        cfg_kwargs.pop("negative_prompt", None)
                        cfg = gtypes.EditImageConfig(**cfg_kwargs)

                    resp = self.client.models.edit_image(
                        model="imagen-3.0-capability-001",
                        prompt=self.input_prompt or "",
                        reference_images=[raw_ref, mask_ref],
                        config=cfg,
                    )
                    imgs = getattr(resp, "generated_images", None) or []
                    for idx, gi in enumerate(imgs[: min(self.num, self.imagen_max_num)]):
                        data = self._extract_imagen_bytes(gi)
                        p = self._save(idx, data)
                        if p:
                            paths.append(p)

                    # store reference for future remix: prefer remote URI if available, otherwise saved path
                    if paths:
                        self._store_image_reference_imagen(imgs[0] if imgs else None, paths[0])

                else:
                    # Gemini Developer API remix via generate_content with prompt + reference image part.
                    ref_part = self._image_part_from_identifier(self.image_id)
                    if not ref_part:
                        raise RuntimeError("Invalid image_id for remix. Provide a valid local path, Files API name, http(s) URL, or gs:// URI.")
                    img_cfg = self._build_gemini_image_config(self.model, self.resolution)
                    resp = self.client.models.generate_content(
                        model=self.model or self.DEFAULT_GEMINI_IMAGE_MODEL,
                        contents=[self.input_prompt or "", ref_part],
                        config=gtypes.GenerateContentConfig(
                            image_config=img_cfg,
                        ),
                    )
                    saved = 0
                    for cand in getattr(resp, "candidates", []) or []:
                        parts = getattr(getattr(cand, "content", None), "parts", None) or []
                        for part in parts:
                            inline = getattr(part, "inline_data", None)
                            if inline and getattr(inline, "data", None):
                                p = self._save(saved, inline.data)
                                if p:
                                    paths.append(p)
                                    saved += 1
                                    if saved >= self.num:
                                        break
                        if saved >= self.num:
                            break

                    # store reference: saved local path is a reusable identifier for next remix
                    if paths:
                        self._store_image_id(paths[0])

                if self.inline:
                    self.signals.finished_inline.emit(self.ctx, paths, self.input_prompt)
                else:
                    self.signals.finished.emit(self.ctx, paths, self.input_prompt)
                return  # remix path finished

            # Normal paths
            self.signals.status.emit(trans('img.status.generating') + f": {self.input_prompt}...")

            if self.mode == Image.MODE_EDIT:
                # EDIT
                if self._using_vertex():
                    # Vertex Imagen edit API (preferred)
                    resp = self._imagen_edit(self.input_prompt, self.attachments, self.num)
                    imgs = getattr(resp, "generated_images", None) or []
                    for idx, gi in enumerate(imgs[: self.num]):
                        data = self._extract_imagen_bytes(gi)
                        p = self._save(idx, data)
                        if p:
                            paths.append(p)
                    # store reference
                    if paths:
                        self._store_image_reference_imagen(imgs[0] if imgs else None, paths[0])
                else:
                    # Gemini Developer API via Gemini image models (Nano Banana / Nano Banana Pro)
                    resp = self._gemini_edit(self.input_prompt, self.attachments, self.num)
                    saved = 0
                    for cand in getattr(resp, "candidates", []) or []:
                        parts = getattr(getattr(cand, "content", None), "parts", None) or []
                        for part in parts:
                            inline = getattr(part, "inline_data", None)
                            if inline and getattr(inline, "data", None):
                                p = self._save(saved, inline.data)
                                if p:
                                    paths.append(p)
                                    saved += 1
                                    if saved >= self.num:
                                        break
                        if saved >= self.num:
                            break
                    # store reference
                    if paths:
                        self._store_image_id(paths[0])

            else:
                # GENERATE
                if self._is_imagen_generate(self.model) and self._using_vertex():
                    num = min(self.num, self.imagen_max_num)
                    resp = self._imagen_generate(self.input_prompt, num, self.resolution)
                    imgs = getattr(resp, "generated_images", None) or []
                    for idx, gi in enumerate(imgs[: num]):
                        data = self._extract_imagen_bytes(gi)
                        p = self._save(idx, data)
                        if p:
                            paths.append(p)
                    # store reference
                    if paths:
                        self._store_image_reference_imagen(imgs[0] if imgs else None, paths[0])
                else:
                    # Gemini Developer API image generation (Nano Banana / Nano Banana Pro) with robust sizing + optional reference images
                    resp = self._gemini_generate_image(self.input_prompt, self.model, self.resolution)
                    saved = 0
                    for cand in getattr(resp, "candidates", []) or []:
                        parts = getattr(getattr(cand, "content", None), "parts", None) or []
                        for part in parts:
                            inline = getattr(part, "inline_data", None)
                            if inline and getattr(inline, "data", None):
                                p = self._save(saved, inline.data)
                                if p:
                                    paths.append(p)
                                    saved += 1
                                    if saved >= self.num:
                                        break
                        if saved >= self.num:
                            break
                    # store reference
                    if paths:
                        self._store_image_id(paths[0])

            if self.inline:
                self.signals.finished_inline.emit(self.ctx, paths, self.input_prompt)
            else:
                self.signals.finished.emit(self.ctx, paths, self.input_prompt)

        except Exception as e:
            self.signals.error.emit(e)
        finally:
            self._cleanup()

    # ---------- helpers ----------

    def _using_vertex(self) -> bool:
        """
        Detect if Vertex AI is configured via env vars.
        """
        val = os.getenv("GOOGLE_GENAI_USE_VERTEXAI") or ""
        return str(val).lower() in ("1", "true", "yes", "y")

    def _is_imagen_generate(self, model_id: str) -> bool:
        """True for Imagen generate models."""
        mid = str(model_id).lower()
        return "imagen" in mid and "generate" in mid

    def _imagen_supports_negative_prompt(self, model_id: str) -> bool:
        """
        Return True if the Imagen model supports native negative_prompt.
        Supported: imagen-3.0-generate-001, imagen-3.0-fast-generate-001, imagen-3.0-capability-001.
        """
        mid = str(model_id or "").lower()
        return any(x in mid for x in (
            "imagen-3.0-generate-001",
            "imagen-3.0-fast-generate-001",
            "imagen-3.0-capability-001",
        ))

    def _imagen_generate(self, prompt: str, num: int, resolution: str):
        """Imagen text-to-image."""
        aspect = self._aspect_from_resolution(resolution)
        # Build config with optional negative_prompt when supported by model and provided.
        cfg_kwargs: Dict[str, Any] = {"number_of_images": num}
        if aspect:
            cfg_kwargs["aspect_ratio"] = aspect
        if self.extra_prompt and self._imagen_supports_negative_prompt(self.model):
            cfg_kwargs["negative_prompt"] = self.extra_prompt
        try:
            cfg = gtypes.GenerateImagesConfig(**cfg_kwargs)
        except Exception:
            # Fallback without negative_prompt if SDK doesn't recognize it
            cfg_kwargs.pop("negative_prompt", None)
            cfg = gtypes.GenerateImagesConfig(**cfg_kwargs)

        return self.client.models.generate_images(
            model=self.model,
            prompt=prompt,
            config=cfg,
        )

    def _imagen_edit(self, prompt: str, attachments: Dict[str, Any], num: int):
        """
        Imagen edit: requires Vertex AI and capability model (e.g. imagen-3.0-capability-001).
        First attachment = base image, optional second = mask.
        """
        paths = self._collect_attachment_paths(attachments)
        if len(paths) == 0:
            raise RuntimeError("No attachment provided for edit mode.")

        base_img = gtypes.Image.from_file(location=paths[0])
        raw_ref = gtypes.RawReferenceImage(reference_id=0, reference_image=base_img)

        if len(paths) >= 2:
            mask_img = gtypes.Image.from_file(location=paths[1])
            mask_ref = gtypes.MaskReferenceImage(
                reference_id=1,
                reference_image=mask_img,
                config=gtypes.MaskReferenceConfig(
                    mask_mode="MASK_MODE_USER_PROVIDED",
                    mask_dilation=0.0,
                ),
            )
            edit_mode = "EDIT_MODE_INPAINT_INSERTION"
        else:
            mask_ref = gtypes.MaskReferenceImage(
                reference_id=1,
                reference_image=None,
                config=gtypes.MaskReferenceConfig(
                    mask_mode="MASK_MODE_BACKGROUND",
                    mask_dilation=0.0,
                ),
            )
            edit_mode = "EDIT_MODE_BGSWAP"

        # Build edit config with optional negative_prompt
        cfg_kwargs = dict(
            edit_mode=edit_mode,
            number_of_images=min(num, self.imagen_max_num),
            include_rai_reason=True,
        )
        if self.extra_prompt and self._imagen_supports_negative_prompt(self.model):
            cfg_kwargs["negative_prompt"] = self.extra_prompt
        try:
            cfg = gtypes.EditImageConfig(**cfg_kwargs)
        except Exception:
            cfg_kwargs.pop("negative_prompt", None)
            cfg = gtypes.EditImageConfig(**cfg_kwargs)

        # Ensure capability model for edit
        model_id = "imagen-3.0-capability-001"
        return self.client.models.edit_image(
            model=model_id,
            prompt=prompt,
            reference_images=[raw_ref, mask_ref],
            config=cfg,
        )

    def _is_gemini_pro_image_model(self, model_id: str) -> bool:
        """
        Detect Gemini 3 Pro Image (Nano Banana Pro) by id or UI alias.
        """
        mid = (model_id or "").lower()
        return mid.startswith("gemini-") or mid.startswith("nano-banana") or mid.startswith("nb-")

    def _infer_nb_pro_size_for_dims(self, w: int, h: int) -> Optional[str]:
        """
        Infer '1K' | '2K' | '4K' for Nano Banana Pro from WxH.
        """
        dims = f"{w}x{h}"
        if dims in self._NB_PRO_1K:
            return "1K"
        if (w % 2 == 0) and (h % 2 == 0):
            if f"{w // 2}x{h // 2}" in self._NB_PRO_1K:
                return "2K"
        if (w % 4 == 0) and (h % 4 == 0):
            if f"{w // 4}x{h // 4}" in self._NB_PRO_1K:
                return "4K"
        mx = max(w, h)
        if mx >= 4000:
            return "4K"
        if mx >= 2000:
            return "2K"
        return "1K"

    def _build_gemini_image_config(self, model_id: str, resolution: str) -> Optional[gtypes.ImageConfig]:
        """
        Build ImageConfig for Gemini image models.
        """
        try:
            aspect = self._aspect_from_resolution(resolution)
            cfg = gtypes.ImageConfig()
            if aspect:
                cfg.aspect_ratio = aspect

            # Only Pro supports image_size; detect by id/alias and set 1K/2K/4K from WxH.
            if self._is_gemini_pro_image_model(model_id):
                w_str, h_str = resolution.lower().replace("×", "x").split("x")
                w, h = int(w_str.strip()), int(h_str.strip())
                k = self._infer_nb_pro_size_for_dims(w, h)
                if k:
                    cfg.image_size = k
            return cfg
        except Exception:
            return None

    def _attachment_image_parts(self) -> List[gtypes.Part]:
        """
        Build image Parts from current attachments for Gemini models.
        """
        parts: List[gtypes.Part] = []
        paths = self._collect_attachment_paths(self.attachments)
        for p in paths:
            try:
                mime = self._guess_mime(p)
                if not mime or not mime.startswith("image/"):
                    continue
                with open(p, "rb") as f:
                    data = f.read()
                parts.append(gtypes.Part.from_bytes(data=data, mime_type=mime))
            except Exception:
                continue
        return parts

    def _gemini_generate_image(self, prompt: str, model_id: str, resolution: str):
        """
        Call Gemini generate_content with robust fallback for image_size.
        Supports optional reference images uploaded as attachments.
        """
        cfg = self._build_gemini_image_config(model_id, resolution)
        image_parts = self._attachment_image_parts()

        def _do_call(icfg: Optional[gtypes.ImageConfig]):
            contents: List[Any] = []
            contents.append(prompt or "")
            if image_parts:
                contents.extend(image_parts)
            return self.client.models.generate_content(
                model=model_id or self.DEFAULT_GEMINI_IMAGE_MODEL,
                contents=contents,
                config=gtypes.GenerateContentConfig(
                    response_modalities=[gtypes.Modality.TEXT, gtypes.Modality.IMAGE],
                    image_config=icfg,
                ),
            )

        try:
            return _do_call(cfg)
        except Exception as e:
            msg = str(e)
            if "imageSize" in msg or "image_size" in msg or "Unrecognized" in msg or "unsupported" in msg:
                try:
                    if cfg and getattr(cfg, "image_size", None):
                        cfg2 = gtypes.ImageConfig()
                        cfg2.aspect_ratio = getattr(cfg, "aspect_ratio", None)
                        return _do_call(cfg2)
                except Exception:
                    pass
            raise

    def _gemini_edit(self, prompt: str, attachments: Dict[str, Any], num: int):
        """
        Gemini image-to-image editing via generate_content.
        The first attachment is used as the input image. Honors aspect_ratio and (for Pro) image_size.
        """
        paths = self._collect_attachment_paths(attachments)
        if len(paths) == 0:
            raise RuntimeError("No attachment provided for edit mode.")

        img_path = paths[0]
        with open(img_path, "rb") as f:
            img_bytes = f.read()
        mime = self._guess_mime(img_path)

        cfg = self._build_gemini_image_config(self.model, self.resolution)

        def _do_call(icfg: Optional[gtypes.ImageConfig]):
            return self.client.models.generate_content(
                model=self.model or self.DEFAULT_GEMINI_IMAGE_MODEL,
                contents=[prompt, gtypes.Part.from_bytes(data=img_bytes, mime_type=mime)],
                config=gtypes.GenerateContentConfig(
                    image_config=icfg,
                ),
            )

        try:
            return _do_call(cfg)
        except Exception as e:
            msg = str(e)
            if "imageSize" in msg or "image_size" in msg or "Unrecognized" in msg or "unsupported" in msg:
                if cfg and getattr(cfg, "image_size", None):
                    cfg2 = gtypes.ImageConfig()
                    cfg2.aspect_ratio = getattr(cfg, "aspect_ratio", None)
                    return _do_call(cfg2)
            raise

    def _image_part_from_identifier(self, identifier: str) -> Optional[gtypes.Part]:
        """
        Build a Gemini Part from a generic image identifier:
        - Local path -> Part.from_bytes
        - Files API name (files/...) -> resolve to URI + mime and use Part.from_uri
        - gs:// URI -> Part.from_uri
        - http(s) URL -> download bytes and use Part.from_bytes
        - data: URI (base64) -> decode and use Part.from_bytes
        """
        if not identifier:
            return None
        ident = str(identifier).strip()

        # Local file
        if os.path.exists(ident):
            mime = self._guess_mime(ident)
            with open(ident, "rb") as f:
                return gtypes.Part.from_bytes(data=f.read(), mime_type=mime)

        # Files API
        if ident.startswith("files/"):
            try:
                f = self.client.files.get(name=ident)
                file_uri = getattr(f, "uri", None)
                mime = getattr(f, "mime_type", None) or self._guess_mime_from_uri(file_uri)
                if file_uri and mime:
                    return gtypes.Part.from_uri(file_uri=file_uri, mime_type=mime)
            except Exception:
                pass

        # gs://
        if ident.startswith("gs://"):
            mime = self._guess_mime_from_uri(ident) or "image/png"
            return gtypes.Part.from_uri(file_uri=ident, mime_type=mime)

        # http(s)
        if ident.startswith("http://") or ident.startswith("https://"):
            try:
                r = requests.get(ident, timeout=60)
                if r.status_code == 200:
                    mime = r.headers.get("Content-Type") or self._guess_mime_from_uri(ident) or "image/png"
                    return gtypes.Part.from_bytes(data=r.content, mime_type=mime)
            except Exception:
                return None

        # data:
        if ident.startswith("data:"):
            try:
                head, b64 = ident.split(",", 1)
                mime = head.split(";")[0][5:] if ";" in head else "image/png"
                return gtypes.Part.from_bytes(data=base64.b64decode(b64), mime_type=mime)
            except Exception:
                return None

        return None

    def _imagen_image_from_identifier(self, identifier: str) -> Optional[gtypes.Image]:
        """
        Build a gtypes.Image for Imagen edit:
        - Local path -> Image.from_file
        - Files API name -> resolve to URI; if gs:// use gcs_uri, otherwise download to temp and from_file
        - gs:// -> Image(gcs_uri=...)
        - http(s) -> download to temp file, then from_file
        """
        if not identifier:
            return None
        ident = str(identifier).strip()

        if os.path.exists(ident):
            return gtypes.Image.from_file(location=ident)

        if ident.startswith("files/"):
            try:
                f = self.client.files.get(name=ident)
                uri = getattr(f, "uri", None)
                if uri and uri.startswith("gs://"):
                    return gtypes.Image(gcs_uri=uri)
                if uri and (uri.startswith("http://") or uri.startswith("https://")):
                    tmp = self._download_to_temp(uri)
                    return gtypes.Image.from_file(location=tmp) if tmp else None
            except Exception:
                return None

        if ident.startswith("gs://"):
            return gtypes.Image(gcs_uri=ident)

        if ident.startswith("http://") or ident.startswith("https://"):
            tmp = self._download_to_temp(ident)
            return gtypes.Image.from_file(location=tmp) if tmp else None

        return None

    def _download_to_temp(self, url: str) -> Optional[str]:
        """Download URL to a temporary file and return its path."""
        try:
            r = requests.get(url, timeout=60)
            if r.status_code == 200:
                ext = ".png"
                ct = r.headers.get("Content-Type") or ""
                if "jpeg" in ct:
                    ext = ".jpg"
                elif "webp" in ct:
                    ext = ".webp"
                fd, path = tempfile.mkstemp(suffix=ext)
                with os.fdopen(fd, "wb") as f:
                    f.write(r.content)
                return path
        except Exception:
            return None
        return None

    def _collect_attachment_paths(self, attachments: Dict[str, Any]) -> List[str]:
        """Extract file paths from attachments dict."""
        out: List[str] = []
        for _, att in (attachments or {}).items():
            try:
                if getattr(att, "path", None) and os.path.exists(att.path):
                    out.append(att.path)
            except Exception:
                continue
        return out

    def _aspect_from_resolution(self, resolution: str) -> Optional[str]:
        """Derive aspect ratio from WxH across supported set."""
        try:
            from math import gcd
            tolerance = 0.08
            w_str, h_str = resolution.lower().replace("×", "x").split("x")
            w, h = int(w_str.strip()), int(h_str.strip())
            if w <= 0 or h <= 0:
                return None
            supported = {
                "1:1": 1 / 1,
                "2:3": 2 / 3,
                "3:2": 3 / 2,
                "3:4": 3 / 4,
                "4:3": 4 / 3,
                "4:5": 4 / 5,
                "5:4": 5 / 4,
                "9:16": 9 / 16,
                "16:9": 16 / 9,
                "21:9": 21 / 9,
            }
            g = gcd(w, h)
            key = f"{w // g}:{h // g}"
            if key in supported:
                return key
            r = w / h
            best = min(supported.keys(), key=lambda k: abs(r - supported[k]))
            rel_err = abs(r - supported[best]) / supported[best]
            return best if rel_err <= tolerance else None
        except Exception:
            return None

    def _extract_imagen_bytes(self, generated_image) -> Optional[bytes]:
        """Extract bytes from Imagen GeneratedImage."""
        img = getattr(generated_image, "image", None)
        if not img:
            return None
        data = getattr(img, "image_bytes", None)
        if isinstance(data, (bytes, bytearray)):
            return bytes(data)
        if isinstance(data, str):
            try:
                return base64.b64decode(data)
            except Exception:
                return None
        url = getattr(img, "url", None) or getattr(img, "uri", None)
        if url:
            try:
                r = requests.get(url, timeout=30)
                if r.status_code == 200:
                    return r.content
            except Exception:
                pass
        return None

    def _store_image_reference_imagen(self, generated_image_item: Any, fallback_path: Optional[str]) -> None:
        """
        Persist a reusable image reference to ctx.extra['image_id'].
        Prefer remote URI/name if provided by Imagen; fallback to the saved local path.
        """
        ref = None
        try:
            img = getattr(generated_image_item, "image", None) if generated_image_item else None
            if img:
                ref = getattr(img, "uri", None) or getattr(img, "url", None) or getattr(img, "name", None)
        except Exception:
            ref = None
        self._store_image_id(ref or fallback_path)

    def _store_image_id(self, value: Optional[str]) -> None:
        """
        Store image_id reference in ctx.extra and persist the context item.
        """
        if not value:
            return
        try:
            if not isinstance(self.ctx.extra, dict):
                self.ctx.extra = {}
            self.ctx.extra["image_id"] = str(value)
            self.window.core.ctx.update_item(self.ctx)
        except Exception:
            pass

    def _save(self, idx: int, data: Optional[bytes]) -> Optional[str]:
        """Save image bytes to file and return path."""
        if not data:
            return None
        name = (
            datetime.date.today().strftime("%Y-%m-%d") + "_" +
            datetime.datetime.now().strftime("%H-%M-%S") + "-" +
            self.window.core.image.make_safe_filename(self.input_prompt) + "-" +
            str(idx + 1) + ".png"
        )
        path = os.path.join(self.window.core.config.get_user_dir("img"), name)
        self.signals.status.emit(trans('img.status.downloading') + f" ({idx + 1} / {self.num}) -> {path}")
        if self.window.core.image.save_image(path, data):
            return path
        return None

    def _guess_mime(self, path: str) -> str:
        """
        Guess MIME type for a local image file.
        """
        mime, _ = mimetypes.guess_type(path)
        if mime:
            return mime
        ext = os.path.splitext(path.lower())[1]
        if ext in ('.jpg', '.jpeg'):
            return 'image/jpeg'
        if ext == '.webp':
            return 'image/webp'
        if ext in ('.heic', '.heif'):
            return 'image/heic'
        return 'image/png'

    def _guess_mime_from_uri(self, uri: Optional[str]) -> Optional[str]:
        """Best-effort MIME guess from URI or file extension."""
        if not uri:
            return None
        mime, _ = mimetypes.guess_type(uri)
        return mime or None

    def _cleanup(self):
        """Cleanup resources."""
        sig = self.signals
        self.signals = None
        if sig is not None:
            try:
                sig.deleteLater()
            except RuntimeError:
                pass

    # ---------- prompt utilities ----------

    @staticmethod
    def _merge_negative_prompt(prompt: str, negative: Optional[str]) -> str:
        """
        Append a negative prompt to the main text prompt when the provider has no native negative_prompt field.
        """
        base = (prompt or "").strip()
        neg = (negative or "").strip()
        if not neg:
            return base
        return (base + ("\n" if base else "") + f"Negative prompt: {neg}").strip()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.01 15:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QCheckBox, QWidget, QPushButton
from PySide6.QtGui import QIcon
from PySide6.QtCore import QSize, Qt

from pygpt_net.ui.base.flow_layout import FlowLayout
from pygpt_net.utils import trans

class OptionCheckboxList(QWidget):
    def __init__(
            self,
            window=None,
            parent_id: str = None,
            id: str = None,
            option: dict = None,
            icon = None
    ):
        """
        Settings checkbox list

        :param window: main window
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        :param icon: icon
        """
        super(OptionCheckboxList, self).__init__(window)
        self.window = window
        self.id = id
        self.parent_id = parent_id
        self.option = option
        self.value = False
        self.title = ""
        self.real_time = False
        self.keys = {}
        self.boxes = {}

        # overlay button config
        self._overlay_margin = 4  # px
        self.btn_select = None

        # init from option data
        if self.option is not None:
            if "label" in self.option and self.option["label"] is not None \
                    and self.option["label"] != "":
                self.title = trans(self.option["label"])
            if "value" in self.option:
                self.value = self.option["value"]
            if "real_time" in self.option:
                self.real_time = self.option["real_time"]
            if "keys" in self.option:
                self.keys = self.option["keys"]

        values = self.value.split(",") if isinstance(self.value, str) else []
        widgets = []

        # create a checkbox for each key
        for item in self.keys:
            # item is a dict with key:name
            if type(item) is dict:
                for key, name in item.items():
                    self.boxes[key] = QCheckBox(name, self.window)
                    self.boxes[key].setStyleSheet("margin-left: 5px; margin-right: 5px;")
                    if key in values:
                        self.boxes[key].setChecked(True)
                    else:
                        self.boxes[key].setChecked(False)
                    self.boxes[key].stateChanged.connect(
                        lambda state, k=key: self.window.controller.config.checkbox_list.on_update(
                            self.parent_id,
                            self.id,
                            self.option,
                            state,
                            k
                        )
                    )
                    widgets.append(self.boxes[key])

        self.layout = FlowLayout()
        for widget in widgets:
            self.layout.addWidget(widget)

        # do not add the select/unselect all button to the flow layout
        # it will be overlaid in the top-right corner to avoid affecting layout flow
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        # overlay select/unselect all button pinned to top-right corner
        self.btn_select = QPushButton(self)
        self.btn_select.setObjectName("btn_select_all_overlay")
        self.btn_select.setToolTip(trans("action.select_unselect_all"))
        self.btn_select.setIcon(QIcon(":/icons/asterisk.svg"))
        self.btn_select.setIconSize(QSize(16, 16))
        self.btn_select.setFixedSize(22, 22)
        self.btn_select.setFocusPolicy(Qt.NoFocus)
        self.btn_select.setCursor(Qt.PointingHandCursor)
        self.btn_select.setStyleSheet("QPushButton { border: none; padding: 0; background: transparent; }")
        self.btn_select.clicked.connect(
            lambda: self.window.controller.config.checkbox_list.on_select_all(
                self.parent_id,
                self.id,
                self.option
            )
        )
        self.btn_select.raise_()
        self._place_select_button()

    def _place_select_button(self) -> None:
        """
        Place the overlay select/unselect button in the top-right corner.
        """
        if not self.btn_select:
            return
        m = self._overlay_margin
        w = self.btn_select.width()
        x = max(0, self.width() - w - m)
        y = m
        self.btn_select.move(x, y)

    def resizeEvent(self, event):
        """
        Keep the overlay button pinned to the top-right on resize.
        """
        super().resizeEvent(event)
        self._place_select_button()

    def update_boxes_list(self, keys: dict) -> None:
        """
        Update checkbox list with new keys

        :param keys: dictionary with keys and names
        """
        self.keys = keys

        # remove all existing checkboxes from the layout
        while self.layout.count() > 0:
            item = self.layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.setParent(None)
                widget.deleteLater()

        # clear the boxes dictionary
        self.boxes.clear()

        # get current values from the option
        values = self.value.split(",") if isinstance(self.value, str) else []

        # create a checkbox for each key
        for item in self.keys:
            # item is a dict with key:name
            if type(item) is dict:
                for key, name in item.items():
                    checkbox = QCheckBox(name, self.window)
                    checkbox.setStyleSheet("margin-left: 5px; margin-right: 5px;")
                    checkbox.setChecked(key in values)
                    checkbox.stateChanged.connect(
                        lambda state, k=key: self.window.controller.config.checkbox_list.on_update(
                            self.parent_id,
                            self.id,
                            self.option,
                            state,
                            k
                        )
                    )
                    self.boxes[key] = checkbox
                    self.layout.addWidget(checkbox)

        # keep the overlay button in place after list update
        self._place_select_button()

    def setIcon(self, icon: str):
        """
        Set icon

        :param icon: icon
        """
        pass

    def setText(self, key: str, text: str):
        """
        Set label text

        :param key: checkbox key
        :param text: text
        """
        if key in self.boxes:
            self.boxes[key].setText(text)

    def setChecked(self, key: str, state: bool):
        """
        Set checkbox state

        :param key: checkbox key
        :param state: state
        """
        if key in self.boxes:
            self.boxes[key].setChecked(state)
            # on update hooks
            self.window.controller.config.checkbox_list.on_update(
                self.parent_id,
                self.id,
                self.option,
                state,
                key
            )

    def isChecked(self, key: str) -> bool:
        """
        Get checkbox state

        :return: state
        """
        if key in self.boxes:
            return self.boxes[key].isChecked()
        return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.04 00:00:00                  #
# ================================================== #

import json
from dataclasses import dataclass
from typing import Optional, Dict, Any, ClassVar

from pygpt_net.item.ctx import CtxItem


@dataclass(slots=True)
class BaseEvent:
    """Base Event object class"""
    id: ClassVar[str] = None
    name: Optional[str] = None
    data: Optional[dict] = None
    ctx: Optional[CtxItem] = None  # CtxItem
    stop: bool = False  # True to stop propagation
    internal: bool = False
    call_id: int = 0

    def __post_init__(self):
        # Normalize None to empty dict for convenience and safety
        if self.data is None:
            self.data = {}

    @property
    def full_name(self) -> str:
        """
        Get full event name

        :return: Full event name
        """
        return self.id + ": " + self.name  # type: ignore[operator]

    def to_dict(self) -> Dict[str, Any]:
        """Dump event to dict"""
        return {
            'name': self.name,
            'data': self.data,
            'stop': self.stop,
            'internal': self.internal,
        }

    def dump(self) -> str:
        """
        Dump event to json string

        :return: JSON string
        """
        try:
            return json.dumps(self.to_dict())
        except Exception:
            pass
        return ""

    def __str__(self):
        return self.dump()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.13 16:00:00                  #
# ================================================== #

import os
import configparser
from typing import Optional, Dict, Tuple

from pygpt_net.config import Config


class Locale:
    def __init__(
            self,
            domain: Optional[str] = None,
            config: Optional[Config] = None
    ):
        """
        Locale loader

        :param domain: translation domain
        :param config: Config instance
        """
        self.config = config if config is not None else Config()
        self.lang = 'en'  # default language = en
        self.fallback = 'en'  # fallback language = en
        self.default_domain = 'locale'  # default translation domain
        self.ini_key = 'LOCALE'  # ini key for translations
        self.data: Dict[str, Dict[str, str]] = {}

        # cache: path -> (mtime, parsed_dict)
        self._file_cache: Dict[str, Tuple[float, Dict[str, str]]] = {}

        # load config once
        self.config.init(False)
        if self.config.has('lang'):
            self.lang = self.config.get_lang()

        self.load(self.lang, domain)

    def _clear_cache(self):
        """Clear internal file cache."""
        self._file_cache.clear()

    def reload_config(self):
        """Reload configuration"""
        workdir = self.config.prepare_workdir()
        self.config.set_workdir(workdir)
        self.config.load(False)
        if self.config.has('lang'):
            self.lang = self.config.get_lang()
        self._clear_cache()  # ensure fresh read
        self.load(self.lang)

    def reload(self, domain: Optional[str] = None):
        """
        Reload translations for domain

        :param domain: translation domain
        """
        self.config.load(False)
        if self.config.has('lang'):
            self.lang = self.config.get_lang()
        self._clear_cache()  # ensure fresh read
        self.load(self.lang, domain)

    def from_file(self, path: str) -> dict:
        """
        Load and parse translations from file (cached by mtime)

        :param path: path to ini file
        :return: dict with translations
        """
        try:
            mtime = os.path.getmtime(path)
        except FileNotFoundError:
            return {}

        cached = self._file_cache.get(path)
        if cached is not None and cached[0] == mtime:
            return cached[1]

        # RawConfigParser with interpolation disabled
        ini = configparser.RawConfigParser(interpolation=None)
        ini.read(path, encoding='utf-8')

        data = dict(ini.items(self.ini_key))
        self._file_cache[path] = (mtime, data)
        return data

    def load_by_lang(
            self,
            lang: str,
            domain: Optional[str] = None
    ):
        """
        Load translation data by language code

        :param lang: language code
        :param domain: translation domain
        """
        domain_id = domain or self.default_domain
        mapping = self.data.setdefault(domain_id, {})

        try:
            base_path = self.get_base_path(domain_id, lang)
            if os.path.isfile(base_path):
                mapping.update(self.from_file(base_path))

            user_path = self.get_user_path(domain_id, lang)
            if os.path.isfile(user_path):
                mapping.update(self.from_file(user_path))
        except Exception as e:
            print(e)

    def load(
            self,
            lang: str,
            domain: Optional[str] = None
    ):
        """
        Load translation data

        :param lang: language code
        :param domain: translation domain
        """
        if not isinstance(lang, str) or not lang:
            lang = self.fallback

        if lang != self.fallback:
            self.load_by_lang(self.fallback, domain)
        self.load_by_lang(lang, domain)

    def get(
            self,
            key: str,
            domain: Optional[str] = None,
            params: Optional[dict] = None
    ) -> str:
        """
        Return translation for key and domain

        :param key: translation key
        :param domain: translation domain
        :param params: translation params dict for replacement
        :return: translated string or key if not found
        """
        domain_id = domain or self.default_domain
        mapping = self.data.get(domain_id)

        if mapping is None:
            self.load(self.lang, domain)
            mapping = self.data.get(domain_id, {})

        val = mapping.get(key)
        if val is None:
            return key
        text = val.replace('\\n', '\n')

        if isinstance(params, dict) and params:
            try:
                return text.format(**params)
            except KeyError:
                pass
        return text

    def get_base_path(
            self,
            domain: str,
            lang: str
    ) -> str:
        """
        Get base path for locale file

        :param domain: translation domain
        :param lang: language code
        :return: path to translations file
        """
        return os.path.join(
            self.config.get_app_path(),
            'data', 'locale',
            f'{domain}.{lang}.ini'
        )

    def get_user_path(
            self,
            domain: str,
            lang: str
    ) -> str:
        """
        Get user path for locale file (overwrites base path)

        :param domain: translation domain
        :param lang: language code
        :return: path to translations file
        """
        return os.path.join(
            self.config.get_user_path(),
            'locale',
            f'{domain}.{lang}.ini'
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.26 13:00:00                  #
# ================================================== #

from typing import Any, Dict


class Textarea:
    def __init__(self, window=None):
        """
        Textarea field handler

        :param window: Window instance
        """
        self.window = window

    def apply(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ):
        """
        Apply value to textarea

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        """
        parent = self.window.ui.config[parent_id]
        field = parent.get(key)
        if field is None:
            return
        new_text = str(option["value"]) if "value" in option else ""
        if hasattr(field, "toPlainText"):
            current = field.toPlainText()
        elif hasattr(field, "text"):
            current = field.text()
        else:
            current = None
        if current == new_text:
            return
        if hasattr(field, "setText"):
            field.setText(new_text)
        elif hasattr(field, "setPlainText"):
            field.setPlainText(new_text)

    def on_update(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any],
            value: Any,
            hooks: bool = True
    ):
        """
        Event: on update

        :param parent_id: Parent ID
        :param key: Option key
        :param option: Option data
        :param value: Option value
        :param hooks: Run hooks
        """
        option["value"] = value
        self.apply(parent_id, key, option)

        if hooks:
            hook_name = f"update.{parent_id}.{key}"
            if self.window.ui.has_hook(hook_name):
                hook = self.window.ui.get_hook(hook_name)
                try:
                    hook(key, value, "textarea")
                except Exception as e:
                    self.window.core.debug.log(e)

    def get_value(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ) -> str:
        """
        Get value of textarea

        :param parent_id: Parent ID
        :param key: Option key
        :param option: Option data
        :return: Textarea option value (plain text)
        """
        return self.window.ui.config[parent_id][key].toPlainText()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.23 07:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        dockerfile = 'FROM python:3.9-alpine'
        dockerfile += '\n\n'
        dockerfile += 'RUN mkdir /data'
        dockerfile += '\n\n'
        dockerfile += '# Data directory, bound as a volume to the local \'data/\' directory'
        dockerfile += '\nWORKDIR /data'

        volumes_keys = {
            "enabled": "bool",
            "docker": "text",
            "host": "text",
        }
        volumes_items = [
            {
                "enabled": True,
                "docker": "/data",
                "host": "{workdir}",
            },
        ]
        ports_keys = {
            "enabled": "bool",
            "docker": "text",
            "host": "int",
        }
        ports_items = []

        # Sandbox / sys_exec (original)
        plugin.add_option(
            "sandbox_docker",
            type="bool",
            value=False,
            label="Sandbox (docker container)",
            description="Executes commands in sandbox (docker container). "
                        "Docker must be installed and running.",
            tab="sandbox",
        )
        plugin.add_option(
            "dockerfile",
            type="textarea",
            value=dockerfile,
            label="Dockerfile",
            description="Dockerfile",
            tooltip="Dockerfile",
            tab="sandbox",
        )
        plugin.add_option(
            "image_name",
            type="text",
            value='pygpt_system',
            label="Docker image name",
            tab="sandbox",
        )
        plugin.add_option(
            "container_name",
            type="text",
            value='pygpt_system_container',
            label="Docker container name",
            tab="sandbox",
        )
        plugin.add_option(
            "docker_entrypoint",
            type="text",
            value='tail -f /dev/null',
            label="Docker run command",
            tab="sandbox",
            advanced=True,
        )
        plugin.add_option(
            "docker_volumes",
            type="dict",
            value=volumes_items,
            label="Docker volumes",
            description="Docker volumes mapping",
            tooltip="Docker volumes mapping",
            keys=volumes_keys,
            tab="sandbox",
            advanced=True,
        )
        plugin.add_option(
            "docker_ports",
            type="dict",
            value=ports_items,
            label="Docker ports",
            description="Docker ports mapping",
            tooltip="Docker ports mapping",
            keys=ports_keys,
            tab="sandbox",
            advanced=True,
        )
        plugin.add_option(
            "auto_cwd",
            type="bool",
            value=True,
            label="Auto-append CWD to sys_exec",
            description="Automatically append current working directory to sys_exec command",
            tab="general",
        )
        plugin.add_cmd(
            "sys_exec",
            instruction="execute ANY system command, script or app in user's environment. "
                        "Do not use this command to install Python libraries, use IPython environment and IPython commands instead.",
            params=[
                {
                    "name": "command",
                    "type": "str",
                    "description": "system command",
                    "required": True,
                },
            ],
            enabled=True,
            description="Allows system commands execution",
            tab="general",
        )

        # -------------------------------
        # WinAPI options (new tab)
        # -------------------------------
        plugin.add_option(
            "winapi_enabled",
            type="bool",
            value=True,
            label="Enable WinAPI",
            description="Enable Windows API features on Microsoft Windows.",
            tab="winapi",
        )
        plugin.add_option(
            "win_keys_per_char_delay_ms",
            type="int",
            value=2,
            label="Keys: per-char delay (ms)",
            description="Delay between characters for win_keys_text.",
            tab="winapi",
        )
        plugin.add_option(
            "win_keys_hold_ms",
            type="int",
            value=50,
            label="Keys: hold (ms)",
            description="Hold duration for modifiers in win_keys_send.",
            tab="winapi",
        )
        plugin.add_option(
            "win_keys_gap_ms",
            type="int",
            value=30,
            label="Keys: gap (ms)",
            description="Gap between normal key taps in win_keys_send.",
            tab="winapi",
        )
        plugin.add_option(
            "win_drag_step_delay_ms",
            type="int",
            value=10,
            label="Drag: step delay (ms)",
            description="Delay between intermediate drag steps in win_drag.",
            tab="winapi",
        )

        # -------------------------------
        # WinAPI commands
        # -------------------------------
        plugin.add_cmd(
            "win_list",
            instruction="List top-level windows. You can filter by title substring.",
            params=[
                {"name": "filter_title", "type": "str", "description": "Substring to filter by window title", "required": False},
                {"name": "visible_only", "type": "bool", "description": "Only visible windows", "required": False},
                {"name": "limit", "type": "int", "description": "Limit number of results", "required": False},
            ],
            enabled=True,
            description="List top-level windows",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_find",
            instruction="Find windows by multiple attributes.",
            params=[
                {"name": "title", "type": "str", "description": "Title or substring", "required": False},
                {"name": "class_name", "type": "str", "description": "Exact class name", "required": False},
                {"name": "exe", "type": "str", "description": "Process executable name", "required": False},
                {"name": "pid", "type": "int", "description": "Process ID", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
                {"name": "visible_only", "type": "bool", "description": "Only visible", "required": False},
            ],
            enabled=True,
            description="Find windows",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_children",
            instruction="List child windows for a given parent HWND.",
            params=[{"name": "hwnd", "type": "int", "description": "Parent window handle", "required": True}],
            enabled=True,
            description="List child windows",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_foreground",
            instruction="Return info about the current foreground window.",
            params=[],
            enabled=True,
            description="Foreground window info",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_rect",
            instruction="Return window rectangle (geometry).",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Get window rect",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_get_state",
            instruction="Return full window state and metadata.",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Get window state",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_focus",
            instruction="Bring a window to foreground.",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Set focus to a window",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_move_resize",
            instruction="Move and/or resize a window. Omit any of x,y,width,height to keep current.",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
                {"name": "x", "type": "int", "description": "Left position", "required": False},
                {"name": "y", "type": "int", "description": "Top position", "required": False},
                {"name": "width", "type": "int", "description": "Width", "required": False},
                {"name": "height", "type": "int", "description": "Height", "required": False},
            ],
            enabled=True,
            description="Move/resize window",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_minimize",
            instruction="Minimize a window.",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Minimize window",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_maximize",
            instruction="Maximize a window.",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Maximize window",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_restore",
            instruction="Restore a minimized or maximized window.",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Restore window",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_close",
            instruction="Close a window (sends WM_CLOSE).",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Close window",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_show",
            instruction="Show a window.",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Show window",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_hide",
            instruction="Hide a window.",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Hide window",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_always_on_top",
            instruction="Toggle window always-on-top (topmost).",
            params=[
                {"name": "topmost", "type": "bool", "description": "True to set, False to clear", "required": True},
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Always on top",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_set_opacity",
            instruction="Set window opacity using layered style. Provide alpha (0..255) OR opacity (0..1).",
            params=[
                {"name": "alpha", "type": "int", "description": "Alpha 0..255", "required": False},
                {"name": "opacity", "type": "float", "description": "Opacity 0..1", "required": False},
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
            ],
            enabled=True,
            description="Set window opacity",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_screenshot",
            instruction="Take a window screenshot and save to a PNG file. Returns absolute path.",
            params=[
                {"name": "hwnd", "type": "int", "description": "Target window handle", "required": False},
                {"name": "title", "type": "str", "description": "Title substring to match", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
                {"name": "path", "type": "str", "description": "Output path (relative to data/ if not absolute)", "required": False},
            ],
            enabled=True,
            description="Window screenshot to PNG",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_area_screenshot",
            instruction="Take a rectangular screenshot of the screen or relative to a window.",
            params=[
                {"name": "x", "type": "int", "description": "Left", "required": True},
                {"name": "y", "type": "int", "description": "Top", "required": True},
                {"name": "width", "type": "int", "description": "Width", "required": True},
                {"name": "height", "type": "int", "description": "Height", "required": True},
                {"name": "hwnd", "type": "int", "description": "Base window (optional)", "required": False},
                {"name": "title", "type": "str", "description": "Base window by title (optional)", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
                {"name": "relative", "type": "bool", "description": "Coords relative to window when hwnd/title provided", "required": False},
                {"name": "path", "type": "str", "description": "Output path", "required": False},
            ],
            enabled=True,
            description="Area screenshot to PNG",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_clipboard_get",
            instruction="Get clipboard text.",
            params=[],
            enabled=True,
            description="Clipboard get",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_clipboard_set",
            instruction="Set clipboard text.",
            params=[{"name": "text", "type": "str", "description": "Text to set", "required": True}],
            enabled=True,
            description="Clipboard set",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_cursor_get",
            instruction="Get cursor position.",
            params=[],
            enabled=True,
            description="Cursor get",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_cursor_set",
            instruction="Set cursor position.",
            params=[
                {"name": "x", "type": "int", "description": "X coordinate", "required": True},
                {"name": "y", "type": "int", "description": "Y coordinate", "required": True},
            ],
            enabled=True,
            description="Cursor set",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_keys_text",
            instruction="Type Unicode text into the active window (uses SendInput).",
            params=[
                {"name": "text", "type": "str", "description": "Text to type", "required": True},
                {"name": "per_char_delay_ms", "type": "int", "description": "Delay between characters in ms", "required": False},
            ],
            enabled=True,
            description="Type text",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_keys_send",
            instruction="Send key combinations, e.g. ['CTRL','ALT','T'] or ['F5'].",
            params=[
                {"name": "keys", "type": "list", "description": "List of key tokens", "required": True},
                {"name": "hold_ms", "type": "int", "description": "Hold duration for modifiers", "required": False},
                {"name": "gap_ms", "type": "int", "description": "Gap between taps", "required": False},
            ],
            enabled=True,
            description="Send key combos",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_click",
            instruction="Click at a screen point or relative to a window.",
            params=[
                {"name": "x", "type": "int", "description": "X coordinate", "required": True},
                {"name": "y", "type": "int", "description": "Y coordinate", "required": True},
                {"name": "button", "type": "str", "description": "left/right/middle", "required": False},
                {"name": "double", "type": "bool", "description": "Double click", "required": False},
                {"name": "hwnd", "type": "int", "description": "Base window", "required": False},
                {"name": "title", "type": "str", "description": "Base window by title", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
                {"name": "relative", "type": "bool", "description": "Coords relative to window", "required": False},
            ],
            enabled=True,
            description="Mouse click",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_drag",
            instruction="Drag from (x1,y1) to (x2,y2), absolute or relative to a window.",
            params=[
                {"name": "x1", "type": "int", "description": "Start X", "required": True},
                {"name": "y1", "type": "int", "description": "Start Y", "required": True},
                {"name": "x2", "type": "int", "description": "End X", "required": True},
                {"name": "y2", "type": "int", "description": "End Y", "required": True},
                {"name": "hwnd", "type": "int", "description": "Base window", "required": False},
                {"name": "title", "type": "str", "description": "Base window by title", "required": False},
                {"name": "exact", "type": "bool", "description": "Exact title match", "required": False},
                {"name": "relative", "type": "bool", "description": "Coords relative to window", "required": False},
                {"name": "steps", "type": "int", "description": "Interpolation steps", "required": False},
                {"name": "hold_ms", "type": "int", "description": "Delay between steps (ms)", "required": False},
            ],
            enabled=True,
            description="Mouse drag",
            tab="winapi",
        )
        plugin.add_cmd(
            "win_monitors",
            instruction="List monitors/screens and their geometry.",
            params=[],
            enabled=True,
            description="List monitors",
            tab="winapi",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.06 01:00:00                  #
# ================================================== #

import base64
import re
import time

class LocalKernel:

    NOT_READY_MSG = "IPython kernel is not initialized... try to restart the kernel with: /restart"

    def __init__(self, plugin = None):
        self.plugin = plugin
        self.client = None
        self.manager = None
        self.initialized = False
        self.signals = None
        self.restarting = False

    def restart_kernel(self) -> bool:
        """
        Restart the IPython kernel.

        :return: True if successful.
        """
        if self.restarting:
            self.log("IPython kernel is already restarting.")
            return False

        self.restarting = True
        if self.initialized:
            self.client.stop_channels()
            self.manager.restart_kernel(now=True)
            self.client = self.manager.client()
            self.client.start_channels()
            self.log("IPython kernel restarted.")
        else:
            self.init()
            self.log("IPython kernel started.")
        self.client.wait_for_ready()
        self.restarting = False
        return True

    def shutdown_kernel(self):
        """Shutdown the IPython kernel."""
        self.client.stop_channels()
        self.manager.shutdown_kernel()

    def init(self, force: bool = False):
        """
        Initialize the IPython kernel client.

        :param force: Force reinitialization.
        """
        from jupyter_client import KernelManager
        if self.initialized and not force:
            return
        self.manager = KernelManager()
        self.manager.start_kernel()
        self.client = self.manager.client()
        self.client.start_channels()
        self.client.wait_for_ready()
        self.log("Connected to local IPython kernel.")
        self.initialized = True
        self.log("IPython kernel is ready.")

    def process_message(self, msg: dict) -> str:
        """
        Process the message from the IPython kernel.

        :param msg: Message from the IPython kernel.
        :return: Processed message.
        """
        msg_type = msg['msg_type']
        content = msg['content']
        if msg_type == 'stream':
            # standard output and error
            return content['text']
        elif msg_type == 'display_data':
            # display data
            data = content['data']
            if 'text/plain' in data:
                return data['text/plain']
        elif msg_type == 'execute_result':
            # execution result
            data = content['data']
            if 'text/plain' in data:
                return data['text/plain']
        elif msg_type == 'error':
            # execution errors
            return "Error executing code:" + "\n".join(content['traceback'])
        return ""

    def end(self, all: bool = False):
        """
        Stop the IPython kernel.

        :param all: Stop the container as well.
        """
        self.client.stop_channels()  # stop the client
        self.manager.shutdown_kernel()

    def check_ready(self):
        """
        Check if the IPython kernel is ready.

        :return: True if the kernel is ready, False otherwise.
        """
        try:
            if self.client is not None:
                self.client.wait_for_ready(timeout=1)
                return True
        except Exception as e:
            self.log(f"Error checking IPython kernel readiness: {e}")
        return False

    def execute(
            self,
            code: str,
            current: bool = False,
            auto_init: bool = False
    ) -> str:
        """
        Execute the code in the IPython kernel.

        :param code: Python code to execute.
        :param current: Use the current kernel.
        :param auto_init: Automatically initialize the kernel if not initialized after error.
        :return: Output from the kernel.
        """
        self.init()
        if not self.initialized:
            self.log("IPython kernel is not initialized.")
            self.send_output(self.NOT_READY_MSG)
            return self.NOT_READY_MSG

        if not current:
            self.restart_kernel()
            time.sleep(1)

        self.log("Executing code: " + str(code)[:100] + "...")

        if not self.check_ready():
            self.log("IPython kernel is not ready.")
            self.send_output(self.NOT_READY_MSG)
            return self.NOT_READY_MSG

        msg_id = self.client.execute(code)
        output = ""
        while True:
            try:
                msg = self.client.get_iopub_msg(timeout=1)
            except:
                break

            if msg['parent_header'].get('msg_id') != msg_id:
                continue

            # receive binary image data
            if msg['msg_type'] in ['display_data', 'execute_result']:
                data = msg['content'].get('data', {})
                if 'image/png' in data:
                    b64_image = data['image/png']
                    binary_image = base64.b64decode(b64_image)
                    self.log("Received binary image data.")
                    if binary_image:
                        path_to_save = self.plugin.make_temp_file_path('png')
                        try:
                            with open(path_to_save, 'wb') as f:
                                f.write(binary_image)
                            self.log(f"Image saved to: {path_to_save}")
                            self.send_output(path_to_save)
                            return str(path_to_save)
                        except Exception as e:
                            self.log(f"Error saving image: {e}")
                            self.send_output(f"Error saving image: {e}")
                            return f"Error saving image: {e}"

            chunk = str(self.process_message(msg))
            if chunk.strip() != "":
                output += chunk
                self.send_output(chunk)

            if (msg['msg_type'] == 'status' and
                    msg['content']['execution_state'] == 'idle'):
                break

        return self.remove_ansi(output).strip()

    def send_output(self, output: str):
        """
        Send the output to the output.

        :param output: Output.
        :return: Output.
        """
        if self.signals is not None:
            self.signals.ipython_output.emit(output)

    def remove_ansi_more(self, text):
        """
        Clean the text from ANSI escape sequences, carriage returns, and progress bars.

        :param text: Text to clean.
        :return: Cleaned text.
        """
        # Remove ANSI escape sequences
        ansi_escape = re.compile(r'''
            \x1B  # ESC
            (?:   # Start of sequence
                [@-Z\\-_]  # ESC [@ to ESC _ (7-bit C1 Control codes)
            |     # or
                \[  # ESC[
                [0-?]*  # Parameter bytes
                [ -/]*  # Intermediate bytes
                [@-~]   # Final byte
            )
        ''', re.VERBOSE)
        text = ansi_escape.sub('', text)

        # Split text into lines
        lines = text.split('\n')

        cleaned_lines = []
        for line in lines:
            # Handle carriage returns - keep only the text after the last '\r'
            if '\r' in line:
                line = line.split('\r')[-1]
            # Optionally, remove progress bar lines by detecting lines containing box-drawing characters
            # Skip the line if it contains box-drawing characters
            if re.search(r'[\u2500-\u257F]', line):
                continue  # Skip this line
            # Append the cleaned line
            cleaned_lines.append(line)
        # Reconstruct the text
        text = '\n'.join(cleaned_lines)
        return text

    def remove_ansi(self, text) -> str:
        """
        Clean the text from ANSI escape sequences.

        :param text: Text to clean.
        :return: Cleaned text.
        """
        ansi_escape = re.compile(
            r'''
            \x1B   # ESC
            (?:    # 7-bit C1 Fe
                [@-Z\\-_]
            |      # or 8-bit C1 Fe
                \[
                [0-?]*   # Parameter bytes
                [ -/]*   # Intermediate bytes
                [@-~]    # Final byte
            )
            ''',
            re.VERBOSE
        )
        return ansi_escape.sub('', text)

    def attach_signals(self, signals):
        """
        Attach signals

        :param signals: signals
        """
        self.signals = signals

    def log(self, msg):
        """
        Log the message.

        :param msg: Message to log.
        """
        print(msg)


import os
from typing import Tuple

import requests
from dotenv import load_dotenv
import json
import base64
from PIL import Image
from io import BytesIO
import io
from urllib.parse import urlparse


load_dotenv(override=True)

BLOCKED_DOMAINS = [
    "maliciousbook.com",
    "evilvideos.com",
    "darkwebforum.com",
    "shadytok.com",
    "suspiciouspins.com",
    "ilanbigio.com",
]


def pp(obj):
    print(json.dumps(obj, indent=4))


def show_image(base_64_image):
    image_data = base64.b64decode(base_64_image)
    image = Image.open(BytesIO(image_data))
    image.show()


def calculate_image_dimensions(base_64_image):
    image_data = base64.b64decode(base_64_image)
    image = Image.open(io.BytesIO(image_data))
    return image.size


def sanitize_message(msg: dict) -> dict:
    """Return a copy of the message with image_url omitted for computer_call_output messages."""
    if msg.get("type") == "computer_call_output":
        output = msg.get("output", {})
        if isinstance(output, dict):
            sanitized = msg.copy()
            sanitized["output"] = {**output, "image_url": "[omitted]"}
            return sanitized
    return msg


def create_response(**kwargs):
    url = f"https://api.openai.com/v1/responses"
    headers = {
        "Authorization": f"Bearer {os.getenv('OPENAI_API_KEY')}",
        "Content-Type": "application/json"
    }

    openai_org = os.getenv("OPENAI_ORG")
    if openai_org:
        headers["Openai-Organization"] = openai_org

    response = requests.post(url, headers=headers, json=kwargs)

    if response.status_code != 200:
        print(f"Error: {response.status_code} {response.text}")

    return response.json()


def check_blocklisted_url(url: str) -> None:
    """Raise ValueError if the given URL (including subdomains) is in the blocklist."""
    hostname = urlparse(url).hostname or ""
    if any(
        hostname == blocked or hostname.endswith(f".{blocked}")
        for blocked in BLOCKED_DOMAINS
    ):
        raise ValueError(f"Blocked URL: {url}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.05 01:00:00                  #
# ================================================== #

import base64
import os
from typing import Optional, Dict, List, Union

from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxItem


class Vision:
    def __init__(self, window=None):
        """
        Vision helpers for Anthropic (image input blocks).

        :param window: Window instance
        """
        self.window = window
        self.attachments: Dict[str, str] = {}
        self.urls: List[str] = []
        self.input_tokens = 0

    def build_blocks(
        self,
        content: Union[str, list],
        attachments: Optional[Dict[str, AttachmentItem]] = None,
    ) -> List[dict]:
        """
        Build image content blocks from local attachments.

        :param content: User message text (unused here)
        :param attachments: Attachments dict (id -> AttachmentItem)
        :return: List of Anthropic content blocks
        """
        blocks: List[dict] = []
        self.attachments = {}
        self.urls = []

        if attachments:
            for id_, attachment in attachments.items():
                if attachment.path and os.path.exists(attachment.path):
                    if self.is_image(attachment.path):
                        mime = self._guess_mime(attachment.path)
                        with open(attachment.path, "rb") as f:
                            data = f.read()
                        b64 = base64.b64encode(data).decode("utf-8")
                        blocks.append({
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": mime,
                                "data": b64,
                            }
                        })
                        self.attachments[id_] = attachment.path
                        attachment.consumed = True

        return blocks

    def is_image(self, path: str) -> bool:
        """
        Check if path looks like an image.

        :param path: File path
        :return: True if path has image file extension
        """
        return path.lower().endswith(('.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif', '.webp'))

    def _guess_mime(self, path: str) -> str:
        """
        Guess mime type from file extension.

        :param path: File path
        :return: MIME type string
        """
        ext = os.path.splitext(path)[1].lower().lstrip(".")
        if ext in ("jpg", "jpeg"):
            return "image/jpeg"
        if ext == "png":
            return "image/png"
        if ext == "gif":
            return "image/gif"
        if ext == "bmp":
            return "image/bmp"
        if ext == "webp":
            return "image/webp"
        if ext == "tiff":
            return "image/tiff"
        return "image/jpeg"

    def append_images(self, ctx: CtxItem):
        """
        Append sent images paths to context for UI/history.

        :param ctx: CtxItem
        """
        images = self.get_attachments()
        if len(images) > 0:
            ctx.images = self.window.core.filesystem.make_local_list(list(images.values()))

    def get_attachments(self) -> Dict[str, str]:
        """
        Return attachments dict (id -> path).

        :return: Attachments dictionary
        """
        return self.attachments

    def get_urls(self) -> List[str]:
        """
        Return image urls (unused).

        :return: List of image URLs
        """
        return self.urls

    def reset_tokens(self):
        """Reset input tokens counter."""
        self.input_tokens = 0

    def get_used_tokens(self) -> int:
        """
        Return input tokens counter.

        :return: Number of input tokens
        """
        return self.input_tokens

    def reset(self):
        """Reset state."""
        self.attachments = {}
        self.urls = []
        self.input_tokens = 0


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.18 03:25:00                  #
# ================================================== #

from __future__ import annotations

import json
import math
import os
import re
import time
from typing import Any, Dict, List, Optional, Tuple, Union

import requests
from PySide6.QtCore import Slot

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals

Number = Union[int, float]
LatLon = Tuple[float, float]


class WorkerSignals(BaseSignals):
    pass


class Worker(BaseWorker):
    """
    OpenStreetMap plugin worker:
    - Geocoding (forward/reverse) via Nominatim
    - Search (alias of geocode)
    - Routing via OSRM (driving/walking/cycling), lightweight by default
    - "Static map": returns openstreetmap.org URL (center/zoom or bbox + optional marker)
    - Utility: OSM site URL, directions URL, single XYZ tile download
    """

    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None
        self.msg = None

    # ---------------------- Core runner ----------------------

    @Slot()
    def run(self):
        try:
            responses = []
            for item in self.cmds:
                if self.is_stopped():
                    break
                try:
                    response = None
                    if item["cmd"] in self.plugin.allowed_cmds and self.plugin.has_cmd(item["cmd"]):

                        if item["cmd"] == "osm_geocode":
                            response = self.cmd_osm_geocode(item)
                        elif item["cmd"] == "osm_reverse":
                            response = self.cmd_osm_reverse(item)
                        elif item["cmd"] == "osm_search":
                            response = self.cmd_osm_search(item)
                        elif item["cmd"] == "osm_route":
                            response = self.cmd_osm_route(item)
                        elif item["cmd"] == "osm_staticmap":
                            response = self.cmd_osm_staticmap(item)
                        elif item["cmd"] == "osm_bbox_map":
                            response = self.cmd_osm_bbox_map(item)
                        elif item["cmd"] == "osm_show_url":
                            response = self.cmd_osm_show_url(item)
                        elif item["cmd"] == "osm_route_url":
                            response = self.cmd_osm_route_url(item)
                        elif item["cmd"] == "osm_tile":
                            response = self.cmd_osm_tile(item)

                        if response is not None:
                            responses.append(response)

                except Exception as e:
                    responses.append(self.make_response(item, self.throw_error(e)))

            if responses:
                self.reply_more(responses)
            if self.msg is not None:
                self.status(self.msg)
        except Exception as e:
            self.error(e)
        finally:
            self.cleanup()

    # ---------------------- HTTP / Config helpers ----------------------

    def _timeout(self) -> int:
        try:
            return int(self.plugin.get_option_value("http_timeout") or 30)
        except Exception:
            return 30

    def _headers(self) -> Dict[str, str]:
        ua = (self.plugin.get_option_value("user_agent") or "").strip()
        if not ua:
            ua = "pygpt-net-osm-plugin/1.0 (+https://pygpt.net)"
        return {
            "User-Agent": ua,
            "Accept": "*/*",
        }

    def _nominatim_base(self) -> str:
        return (self.plugin.get_option_value("nominatim_base") or "https://nominatim.openstreetmap.org").rstrip("/")

    def _osrm_base(self) -> str:
        return (self.plugin.get_option_value("osrm_base") or "https://router.project-osrm.org").rstrip("/")

    def _tile_base(self) -> str:
        return (self.plugin.get_option_value("tile_base") or "https://tile.openstreetmap.org").rstrip("/")

    def _nominatim_common_params(self) -> Dict[str, Any]:
        p = {"format": "jsonv2"}
        email = (self.plugin.get_option_value("contact_email") or "").strip()
        if email:
            p["email"] = email
        lang = (self.plugin.get_option_value("accept_language") or "").strip()
        if lang:
            p["accept-language"] = lang
        return p

    # ---------------------- File / util helpers ----------------------

    def _save_bytes(self, data: bytes, out_path: str) -> str:
        local = self.prepare_path(out_path)
        os.makedirs(os.path.dirname(local), exist_ok=True)
        with open(local, "wb") as fh:
            fh.write(data)
        return local

    def prepare_path(self, path: str) -> str:
        if path in [".", "./"]:
            return self.plugin.window.core.config.get_user_dir("data")
        if self.is_absolute_path(path):
            return path
        return os.path.join(self.plugin.window.core.config.get_user_dir("data"), path)

    def is_absolute_path(self, path: str) -> bool:
        return os.path.isabs(path)

    def _slug(self, s: str, maxlen: int = 80) -> str:
        s = re.sub(r"\s+", "_", (s or "").strip())
        s = re.sub(r"[^a-zA-Z0-9_\-\.]", "", s)
        return (s[:maxlen] or "map").strip("._-") or "map"

    def _now(self) -> int:
        return int(time.time())

    def _add_image(self, path: str):
        try:
            if self.ctx is None:
                return
            if path:
                path = self.plugin.window.core.filesystem.to_workdir(path)
            if not hasattr(self.ctx, "images_before") or self.ctx.images_before is None:
                self.ctx.images_before = []
            if path and path not in self.ctx.images_before:
                self.ctx.images_before.append(path)
        except Exception:
            pass

    # ---------------------- Internal helpers ----------------------

    def _call_cmd(self, cmd: str, params: dict) -> dict:
        fn = getattr(self, f"cmd_{cmd}", None)
        if not callable(fn):
            raise RuntimeError(f"Unknown command: {cmd}")
        return fn({"cmd": cmd, "params": params})

    def _get_payload(self, resp: Any) -> Any:
        if isinstance(resp, dict):
            if resp.get("data") is not None:
                return resp["data"]
            if resp.get("result") is not None:
                return resp["result"]
        return resp

    # ---------------------- Geo helpers ----------------------

    def _is_number(self, x: Any) -> bool:
        try:
            float(x)
            return True
        except Exception:
            return False

    def _parse_point(self, v: Any) -> Optional[LatLon]:
        if v is None:
            return None
        if isinstance(v, (list, tuple)) and len(v) >= 2 and self._is_number(v[0]) and self._is_number(v[1]):
            return float(v[0]), float(v[1])
        if isinstance(v, dict) and "lat" in v and "lon" in v and self._is_number(v["lat"]) and self._is_number(v["lon"]):
            return float(v["lat"]), float(v["lon"])
        if isinstance(v, str):
            parts = [p.strip() for p in v.split(",")]
            if len(parts) >= 2 and self._is_number(parts[0]) and self._is_number(parts[1]):
                return float(parts[0]), float(parts[1])
        return None

    def _validate_latlon(self, lat: float, lon: float) -> Tuple[float, float]:
        if not (math.isfinite(lat) and math.isfinite(lon)):
            raise ValueError("Non-finite coordinate")
        if not (-90.0 <= lat <= 90.0 and -180.0 <= lon <= 180.0):
            raise ValueError("Coordinate out of range")
        return float(lat), float(lon)

    def _resolve_point(self, v: Any) -> LatLon:
        pt = self._parse_point(v)
        if pt:
            return self._validate_latlon(pt[0], pt[1])
        q = str(v).strip()
        res = self._nominatim_search({"q": q, "limit": 1})
        if res:
            lat = float(res[0]["lat"])
            lon = float(res[0]["lon"])
            return self._validate_latlon(lat, lon)
        raise RuntimeError(f"Unable to resolve location: {v}")

    def _bbox_from_radius(self, lat: float, lon: float, radius_m: float) -> Tuple[float, float, float, float]:
        dlat = radius_m / 111320.0
        dlon = radius_m / (111320.0 * max(math.cos(math.radians(lat)), 1e-6))
        return (lon - dlon, lat - dlat, lon + dlon, lat + dlat)

    def _bbox_from_coords(self, coords: List[LatLon]) -> Tuple[float, float, float, float]:
        lats = [c[0] for c in coords]
        lons = [c[1] for c in coords]
        return (min(lons), min(lats), max(lons), max(lats))

    # ---------------------- OSM site URL helpers ----------------------

    def _clamp(self, x: float, a: float, b: float) -> float:
        return max(a, min(b, x))

    def _lat_to_yn(self, lat: float) -> float:
        lat = self._clamp(lat, -85.05112878, 85.05112878)
        rad = math.radians(lat)
        y = math.log(math.tan(math.pi / 4.0 + rad / 2.0))
        return (1.0 - y / math.pi) / 2.0

    def _estimate_zoom_for_bbox(self, minlon: float, minlat: float, maxlon: float, maxlat: float,
                                width: int, height: int) -> int:
        dlon = max(1e-9, maxlon - minlon)
        z_lon = math.log2((width * 360.0) / (256.0 * dlon))
        y1 = self._lat_to_yn(minlat)
        y2 = self._lat_to_yn(maxlat)
        dy = max(1e-12, abs(y2 - y1))
        z_lat = math.log2(height / (256.0 * dy))
        z = int(max(0, min(20, math.floor(min(z_lon, z_lat)) - 1)))
        return z

    def _osm_site_url(self, lat: float, lon: float,
                      zoom: int | None = None,
                      marker: bool = True,
                      layers: str | None = None) -> str:
        z = zoom if zoom is not None else int(self.plugin.get_option_value("map_zoom") or 14)
        z = max(0, min(20, int(z)))
        base = "https://www.openstreetmap.org/"
        qparts = []
        if marker:
            qparts.append(f"mlat={lat:.6f}")
            qparts.append(f"mlon={lon:.6f}")
        if layers:
            qparts.append(f"layers={layers}")
        url = base
        if qparts:
            url += "?" + "&".join(qparts)
        url += f"#map={z}/{lat:.6f}/{lon:.6f}"
        return url

    def _directions_url(self, start: LatLon, end: LatLon, mode: str | None = "car") -> str:
        mode = (mode or "car").lower()
        if mode in ("driving", "car"):
            engine = "fossgis_osrm_car"
        elif mode in ("cycling", "bike", "bicycle"):
            engine = "fossgis_osrm_bike"
        else:
            engine = "fossgis_osrm_foot"
        return f"https://www.openstreetmap.org/directions?engine={engine}&route={start[0]:.6f},{start[1]:.6f};{end[0]:.6f},{end[1]:.6f}"

    def _directions_url_coords(self, coords: List[LatLon], mode: str | None = "car") -> str:
        mode = (mode or "car").lower()
        if mode in ("driving", "car"):
            engine = "fossgis_osrm_car"
        elif mode in ("cycling", "bike", "bicycle"):
            engine = "fossgis_osrm_bike"
        else:
            engine = "fossgis_osrm_foot"
        # route expects 'lat,lon;lat,lon;...' (note: OSM site, not OSRM API order)
        parts = [f"{c[0]:.6f},{c[1]:.6f}" for c in coords]
        return f"https://www.openstreetmap.org/directions?engine={engine}&route=" + ";".join(parts)

    # ---------------------- Nominatim wrappers ----------------------

    def _nominatim_search(self, params: Dict[str, Any]) -> List[dict]:
        base = self._nominatim_base()
        p = self._nominatim_common_params()
        p.update(params or {})
        url = f"{base}/search"
        r = requests.get(url, headers=self._headers(), params=p, timeout=self._timeout())
        if not (200 <= r.status_code < 300):
            raise RuntimeError(f"Nominatim search HTTP {r.status_code}: {r.text[:200]}")
        try:
            data = r.json() or []
        except Exception:
            data = []
        return data if isinstance(data, list) else []

    def _nominatim_reverse(self, lat: float, lon: float, params: Dict[str, Any]) -> dict:
        base = self._nominatim_base()
        p = self._nominatim_common_params()
        p.update({"lat": lat, "lon": lon})
        p.update(params or {})
        url = f"{base}/reverse"
        r = requests.get(url, headers=self._headers(), params=p, timeout=self._timeout())
        if not (200 <= r.status_code < 300):
            raise RuntimeError(f"Nominatim reverse HTTP {r.status_code}: {r.text[:200]}")
        try:
            data = r.json() or {}
        except Exception:
            data = {}
        return data if isinstance(data, dict) else {"data": data}

    # ---------------------- OSRM wrapper ----------------------

    def _osrm_route(self, coords: List[LatLon], profile: str = "driving", opts: Optional[Dict[str, Any]] = None) -> dict:
        if not coords or len(coords) < 2:
            raise RuntimeError("At least 2 coordinates required for routing")

        norm: List[str] = []
        for (lat, lon) in coords:
            lat, lon = self._validate_latlon(lat, lon)
            norm.append(f"{lon:.6f},{lat:.6f}")
        coord_str = ";".join(norm)

        base = self._osrm_base()
        profile = (profile or "driving").lower()
        url = f"{base}/route/v1/{profile}/{coord_str}"

        steps_bool = bool((opts or {}).get("steps", False))
        alternatives_raw = (opts or {}).get("alternatives", 0)
        try:
            alternatives_bool = bool(int(alternatives_raw))
        except Exception:
            alternatives_bool = bool(alternatives_raw)

        params = {
            "overview": (opts or {}).get("overview", "false"),
            "geometries": (opts or {}).get("geometries", "polyline6"),
            "steps": "true" if steps_bool else "false",
            "alternatives": "true" if alternatives_bool else "false",
            "annotations": "false",
        }

        r = requests.get(url, headers=self._headers(), params=params, timeout=self._timeout())
        if not (200 <= r.status_code < 300):
            raise RuntimeError(f"OSRM route HTTP {r.status_code}: {r.text[:200]}")

        data = r.json()
        if (data or {}).get("code") != "Ok":
            raise RuntimeError(f"OSRM error: {(data or {}).get('message') or data}")
        data["_request_url"] = r.request.url
        return data

    # ---------------------- Commands ----------------------

    def cmd_osm_geocode(self, item: dict) -> dict:
        p = item.get("params", {})
        q = p.get("query") or p.get("q")
        if not q:
            return self.make_response(item, "Param 'query' required")

        params: Dict[str, Any] = {"q": q}
        if p.get("limit") is not None:
            params["limit"] = int(p["limit"])
        if p.get("countrycodes"):
            params["countrycodes"] = p["countrycodes"]
        if p.get("viewbox"):
            vb = p["viewbox"]
            if isinstance(vb, (list, tuple)) and len(vb) >= 4:
                params["viewbox"] = ",".join(str(x) for x in vb[:4])
            elif isinstance(vb, str):
                params["viewbox"] = vb
            if p.get("bounded") is not None:
                params["bounded"] = 1 if p.get("bounded") else 0
        if p.get("addressdetails") is not None:
            params["addressdetails"] = 1 if p.get("addressdetails") else 0
        if p.get("polygon_geojson"):
            params["polygon_geojson"] = 1

        if p.get("near") or (p.get("near_lat") is not None and p.get("near_lon") is not None):
            try:
                near = p.get("near") or {"lat": p.get("near_lat"), "lon": p.get("near_lon")}
                latn, lonn = self._resolve_point(near)
                radius = float(p.get("radius_m") or 1000.0)
                minlon, minlat, maxlon, maxlat = self._bbox_from_radius(latn, lonn, radius)
                params["viewbox"] = f"{minlon},{minlat},{maxlon},{maxlat}"
                if p.get("bounded") is not None:
                    params["bounded"] = 1 if p.get("bounded") else 0
            except Exception:
                pass

        try:
            results = self._nominatim_search(params)

            norm = []
            zoom_for_url = int(p.get("zoom") or self.plugin.get_option_value("map_zoom") or 14)
            layers = p.get("layers")
            for r in results:
                latf = float(r.get("lat")) if r.get("lat") else None
                lonf = float(r.get("lon")) if r.get("lon") else None
                item_res = {
                    "display_name": r.get("display_name"),
                    "lat": latf,
                    "lon": lonf,
                    "boundingbox": r.get("boundingbox"),
                    "type": r.get("type"),
                    "class": r.get("class"),
                    "importance": r.get("importance"),
                    "osm_id": r.get("osm_id"),
                    "osm_type": r.get("osm_type"),
                    "_raw": r,
                }
                if (latf is not None) and (lonf is not None):
                    item_res["map_url"] = self._osm_site_url(latf, lonf, zoom_for_url, marker=True, layers=layers)
                norm.append(item_res)
            return self.make_response(item, {"query": q, "count": len(norm), "results": norm})
        except Exception as e:
            return self.make_response(item, self.throw_error(e))

    def cmd_osm_reverse(self, item: dict) -> dict:
        p = item.get("params", {})
        lat = p.get("lat")
        lon = p.get("lon")
        pt = self._parse_point([lat, lon]) if (lat is not None and lon is not None) else self._parse_point(p.get("point"))
        if not pt:
            return self.make_response(item, "Params 'lat' and 'lon' or 'point' required")
        lat, lon = pt
        params: Dict[str, Any] = {}
        if p.get("zoom") is not None:
            params["zoom"] = int(p["zoom"])
        if p.get("addressdetails") is not None:
            params["addressdetails"] = 1 if p.get("addressdetails") else 0
        try:
            data = self._nominatim_reverse(lat, lon, params)
            res = {
                "lat": lat, "lon": lon,
                "display_name": data.get("display_name"),
                "boundingbox": data.get("boundingbox"),
                "address": data.get("address"),
                "_raw": data,
            }
            zoom_for_url = int(p.get("zoom") or self.plugin.get_option_value("map_zoom") or 14)
            layers = p.get("layers")
            res["map_url"] = self._osm_site_url(lat, lon, zoom_for_url, marker=True, layers=layers)
            return self.make_response(item, res)
        except Exception as e:
            return self.make_response(item, self.throw_error(e))

    def cmd_osm_search(self, item: dict) -> dict:
        return self._call_cmd("osm_geocode", item.get("params", {}))

    def cmd_osm_route(self, item: dict) -> dict:
        """
        Modes:
          - mode="url": no OSRM call; returns only directions URL + waypoints.
          - mode="summary" (default): OSRM with overview=false, steps=false; returns distance/duration + directions URL.
          - mode="full": OSRM with overview=simplified, geometries=polyline6; optionally include geometry if include_geometry=True.
        """
        p = item.get("params", {})
        try:
            start = self._resolve_point(p.get("start") or [p.get("start_lat"), p.get("start_lon")])
            end = self._resolve_point(p.get("end") or [p.get("end_lat"), p.get("end_lon")])
        except Exception:
            return self.make_response(item, "Params 'start' and 'end' required (address or lat,lon)")

        # Build coords list with optional waypoints
        waypoints_in = p.get("waypoints") or []
        coords: List[LatLon] = [start]
        for w in waypoints_in:
            try:
                coords.append(self._resolve_point(w))
            except Exception:
                continue
        coords.append(end)

        profile = (p.get("profile") or "driving").lower()
        mode = (p.get("mode") or "summary").lower()

        directions_url = self._directions_url_coords(coords, mode=profile)

        result: Dict[str, Any] = {
            "profile": profile,
            "waypoints": [{"lat": c[0], "lon": c[1]} for c in coords],
            "map_url": directions_url,
            "directions_url": directions_url,
        }

        if mode == "url":
            return self.make_response(item, result)

        include_geometry = bool(p.get("include_geometry") or (mode == "full"))
        include_steps = bool(p.get("include_steps") and mode == "full")
        alternatives = p.get("alternatives") or 0

        opts = {
            "overview": "simplified" if include_geometry else "false",
            "geometries": "polyline6",
            "steps": include_steps,
            "alternatives": alternatives,
        }

        try:
            data = self._osrm_route(coords, profile, opts)
            route = (data.get("routes") or [{}])[0]
            result["distance_m"] = route.get("distance")
            result["duration_s"] = route.get("duration")

            if include_geometry and "geometry" in route:
                max_chars = int(p.get("max_polyline_chars") or 5000)
                geom = route.get("geometry")
                if isinstance(geom, str) and len(geom) > max_chars:
                    geom = geom[:max_chars] + "...(truncated)"
                result["geometry_polyline6"] = geom

            if p.get("save_map") or p.get("want_url"):
                preview_bbox = None
                if isinstance(route.get("geometry"), dict) and route["geometry"].get("type") == "LineString":
                    coords_geo: List[LatLon] = [(float(lat), float(lon)) for lon, lat in route["geometry"].get("coordinates", [])]
                    if coords_geo:
                        minlon, minlat, maxlon, maxlat = self._bbox_from_coords(coords_geo)
                        preview_bbox = [minlon, minlat, maxlon, maxlat]

                sm = self._call_cmd("osm_staticmap", {
                    "bbox": preview_bbox,
                    "center": {"lat": start[0], "lon": start[1]} if not preview_bbox else None,
                    "zoom": p.get("zoom"),
                    "markers": [{"lat": start[0], "lon": start[1]}],
                    "layers": p.get("layers"),
                    "width": int(p.get("width") or self._map_width_default()),
                    "height": int(p.get("height") or self._map_height_default()),
                })
                smd = self._get_payload(sm)
                if isinstance(smd, dict) and smd.get("ok") and smd.get("url"):
                    result["preview_url"] = smd["url"]

            if p.get("debug_url"):
                result["osrm_request_url"] = data.get("_request_url")

            return self.make_response(item, result)
        except Exception as e:
            return self.make_response(item, self.throw_error(e))

    def _map_width_default(self) -> int:
        try:
            return int(self.plugin.get_option_value("map_width") or 800)
        except Exception:
            return 800

    def _map_height_default(self) -> int:
        try:
            return int(self.plugin.get_option_value("map_height") or 600)
        except Exception:
            return 600

    def cmd_osm_staticmap(self, item: dict) -> dict:
        """
        Build an openstreetmap.org URL instead of downloading a static image.
        Supports center/zoom or bbox. Uses the first marker (if any) to set ?mlat/mlon.
        """
        p = item.get("params", {})

        width = int(p.get("width") or self._map_width_default())
        height = int(p.get("height") or self._map_height_default())

        center_pt = None
        zoom = None

        if p.get("bbox"):
            vb = p["bbox"]
            try:
                if isinstance(vb, (list, tuple)) and len(vb) >= 4:
                    minlon, minlat, maxlon, maxlat = map(float, vb[:4])
                elif isinstance(vb, str):
                    parts = [float(x) for x in vb.split(",")]
                    minlon, minlat, maxlon, maxlat = parts[:4]
                else:
                    return self.make_response(item, "Invalid 'bbox' format")

                clat = (minlat + maxlat) / 2.0
                clon = (minlon + maxlon) / 2.0
                center_pt = (clat, clon)
                zoom = int(p.get("zoom")) if p.get("zoom") is not None else self._estimate_zoom_for_bbox(
                    minlon, minlat, maxlon, maxlat, width, height
                )
            except Exception:
                return self.make_response(item, "Invalid 'bbox' format")

        if center_pt is None:
            if p.get("center"):
                center_pt = self._resolve_point(p.get("center"))
            elif (p.get("lat") is not None) and (p.get("lon") is not None):
                center_pt = (float(p.get("lat")), float(p.get("lon")))
        if zoom is None:
            z = p.get("zoom")
            zoom = int(z) if z is not None else int(self.plugin.get_option_value("map_zoom") or 14)

        if center_pt is None:
            mpt = None
            markers = p.get("markers") or []
            if isinstance(markers, (list, tuple)) and markers:
                for m in markers:
                    mpt = self._parse_point(m)
                    if mpt:
                        break
            elif isinstance(markers, str):
                mpt = self._parse_point(markers)
            if mpt:
                center_pt = mpt
            else:
                return self.make_response(item, "Param 'center' or 'lat'/'lon' or 'bbox' required")

        lat, lon = center_pt

        mlat = None
        mlon = None
        markers = p.get("markers") or []
        if isinstance(markers, (list, tuple)) and markers:
            for m in markers:
                mpt = self._parse_point(m)
                if mpt:
                    mlat, mlon = mpt[0], mpt[1]
                    break
        elif isinstance(markers, str):
            mpt = self._parse_point(markers)
            if mpt:
                mlat, mlon = mpt[0], mpt[1]
        elif p.get("marker"):
            mlat, mlon = lat, lon

        layers = p.get("layers")
        qparts = []
        if mlat is not None and mlon is not None:
            qparts.append(f"mlat={mlat:.6f}")
            qparts.append(f"mlon={mlon:.6f}")
        if layers:
            qparts.append(f"layers={layers}")

        base = "https://www.openstreetmap.org/"
        url = base
        if qparts:
            url += "?" + "&".join(qparts)
        url += f"#map={int(zoom)}/{lat:.6f}/{lon:.6f}"

        return self.make_response(item, {
            "ok": True,
            "url": url,
            "center": {"lat": lat, "lon": lon},
            "zoom": int(zoom),
            "marker": {"lat": mlat, "lon": mlon} if (mlat is not None and mlon is not None) else None,
            "service": "openstreetmap.org",
        })

    def cmd_osm_bbox_map(self, item: dict) -> dict:
        p = item.get("params", {})
        if not p.get("bbox"):
            return self.make_response(item, "Param 'bbox' required (minlon,minlat,maxlon,maxlat)")
        return self._call_cmd("osm_staticmap", p)

    def cmd_osm_show_url(self, item: dict) -> dict:
        p = item.get("params", {})
        pt = self._parse_point(p.get("point") or [p.get("lat"), p.get("lon")])
        if not pt:
            return self.make_response(item, "Param 'point' or 'lat'+'lon' required")
        lat, lon = pt
        zoom = int(p.get("zoom") or self.plugin.get_option_value("map_zoom") or 15)
        url = self._osm_site_url(lat, lon, zoom=zoom, marker=True, layers=p.get("layers"))
        return self.make_response(item, {"url": url, "lat": lat, "lon": lon, "zoom": zoom})

    def cmd_osm_route_url(self, item: dict) -> dict:
        p = item.get("params", {})
        try:
            start = self._resolve_point(p.get("start") or [p.get("start_lat"), p.get("start_lon")])
            end = self._resolve_point(p.get("end") or [p.get("end_lat"), p.get("end_lon")])
        except Exception:
            return self.make_response(item, "Params 'start' and 'end' required (address or lat,lon)")
        url = self._directions_url(start, end, mode=(p.get("mode") or p.get("profile") or "car"))
        return self.make_response(item, {"url": url})

    def cmd_osm_tile(self, item: dict) -> dict:
        p = item.get("params", {})
        try:
            z = int(p.get("z"))
            x = int(p.get("x"))
            y = int(p.get("y"))
        except Exception:
            return self.make_response(item, "Params 'z','x','y' required (tile indices)")
        url = f"{self._tile_base()}/{z}/{x}/{y}.png"
        r = requests.get(url, headers=self._headers(), timeout=self._timeout())
        if not (200 <= r.status_code < 300):
            return self.make_response(item, {"ok": False, "status": r.status_code, "error": r.text})
        out = p.get("out") or os.path.join("openstreetmap", f"tile_{z}_{x}_{y}.png")
        local = self._save_bytes(r.content, out)
        self._add_image(local)
        return self.make_response(item, {"ok": True, "file": local, "url": url})


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.25 14:00:00                  #
# ================================================== #

from __future__ import annotations
import json
import re
from dataclasses import dataclass
from typing import List, Optional, Tuple, Any, Dict


@dataclass
class RouteDecision:
    route: Optional[str]  # next node id or "end"
    content: str          # human-facing content to display
    raw: str              # raw model output
    valid: bool
    error: Optional[str] = None


def build_router_instruction(
    agent_name: str,
    current_id: str,
    allowed_routes: List[str],
    friendly_map: Dict[str, Any],
) -> str:
    """
    Builds an instruction that forces the model to output JSON with next route and content.

    Additionally, if the provided friendly_map contains role information for any of the
    allowed routes (e.g. friendly_map[id] is a dict with a "role" key), these roles
    are included in the instruction to improve routing. This is optional and included
    only when present and non-empty.
    """
    allowed = ", ".join(allowed_routes)

    # Normalize human-friendly names: accept either a plain string or a dict with
    # common name keys such as "name", "title" or "label".
    def _extract_name(val: Any, default_name: str) -> str:
        if isinstance(val, str):
            return val
        if isinstance(val, dict):
            return str(val.get("name") or val.get("title") or val.get("label") or default_name)
        return default_name

    # Extract names and optional roles for allowed routes, without changing input map semantics.
    friendly_names = {rid: _extract_name(friendly_map.get(rid), rid) for rid in allowed_routes}

    # Roles are optional; include only non-empty strings.
    friendly_roles: Dict[str, str] = {}
    for rid in allowed_routes:
        val = friendly_map.get(rid)
        role_val: Optional[str] = None
        if isinstance(val, dict):
            role_val = val.get("role")
        if isinstance(role_val, str) and role_val.strip():
            friendly_roles[rid] = role_val.strip()

    # Base instruction
    instr = (
        "You are a routing-capable agent in a multi-agent flow.\n"
        f"Your id is: {current_id}, name: {agent_name}.\n"
        "You MUST respond ONLY with a single JSON object and nothing else.\n"
        "Schema:\n"
        '{\n'
        '  "route": "<ID of the next agent from allowed_routes OR the string \'end\'>",\n'
        '  "content": "<final response text for the user (or tool result)>"\n'
        '}\n'
        "Rules:\n"
        f"- allowed_routes: [{allowed}]\n"
        "- If you want to finish the flow, set route to \"end\".\n"
        "- content must contain the user-facing answer (you may include structured data as JSON or Markdown inside content).\n"
        "- Do NOT add any commentary outside of the JSON. No leading or trailing text.\n"
        "- If using tools, still return the final JSON with tool results summarized in content.\n"
        f"- Human-friendly route names: {json.dumps(friendly_names)}\n"
    )

    # Append roles only if any are present (optional).
    if friendly_roles:
        instr += f"- Human-friendly route roles (optional): {json.dumps(friendly_roles)}\n"

    return instr


def _extract_json_block(text: str) -> Optional[str]:
    """
    Extract JSON block from the text. Supports raw JSON or fenced ```json blocks.
    """
    # fenced block
    m = re.search(r"```json\s*(\{.*?\})\s*```", text, flags=re.DOTALL | re.IGNORECASE)
    if m:
        return m.group(1).strip()

    # first JSON object occurrence fallback
    brace_idx = text.find("{")
    if brace_idx != -1:
        # naive attempt to find matching closing brace
        snippet = text[brace_idx:]
        # try to trim trailing content after last closing brace
        last_close = snippet.rfind("}")
        if last_close != -1:
            return snippet[: last_close + 1].strip()
    return None


def parse_route_output(raw_text: str, allowed_routes: List[str]) -> RouteDecision:
    """
    Parse model output enforcing the {route, content} schema.
    """
    text = raw_text.strip()
    candidate = text

    # Try direct JSON parse
    parsed: Optional[Any] = None
    for attempt in range(2):
        try:
            parsed = json.loads(candidate)
            break
        except Exception:
            if attempt == 0:
                block = _extract_json_block(text)
                if block:
                    candidate = block
                    continue
            parsed = None
            break

    if isinstance(parsed, dict):
        route = parsed.get("route")
        content = parsed.get("content", "")
        # Normalize route strings
        if isinstance(route, str):
            route_norm = route.strip().lower()
            if route_norm == "end":
                return RouteDecision(route="end", content=str(content), raw=raw_text, valid=True)
            # exact match against allowed
            if route in allowed_routes:
                return RouteDecision(route=route, content=str(content), raw=raw_text, valid=True)
        return RouteDecision(
            route=None,
            content=str(parsed.get("content", "")),
            raw=raw_text,
            valid=False,
            error=f"Invalid or disallowed route: {route}",
        )

    # Not a valid JSON – fallback, pass through content as-is
    return RouteDecision(route=None, content=raw_text, raw=raw_text, valid=False, error="Malformed JSON")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.05 18:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QPushButton, QHBoxLayout, QVBoxLayout, QScrollArea, QWidget, QSizePolicy

from pygpt_net.ui.base.config_dialog import BaseConfigDialog
from pygpt_net.ui.widget.dialog.editor import EditorDialog
from pygpt_net.utils import trans


class Dictionary(BaseConfigDialog):
    def __init__(self, window=None, *args, **kwargs):
        super(Dictionary, self).__init__(window, *args, **kwargs)
        """
        Dictionary list item editor dialog

        :param window: Window instance
        """
        self.window = window
        self.id = "dictionary"
        self.dialog_id = "dictionary"
        self.dicts = {}
        self.keys = {}
        self.parents = {}
        self.current_idx = None

    def register(self, id: str, key: str, parent: str, option: dict):
        """
        Register dictionary editor options

        :param id: Dictionary ID
        :param key: Option key
        :param parent: Parent ID
        :param option: Dictionary keys
        """
        self.dicts[id] = option
        self.keys[id] = key
        self.parents[id] = parent

    def setup(self):
        """Setup dictionary editor dialogs"""
        ui = self.window.ui
        controller = self.window.controller
        save_text = trans("dialog.preset.btn.save")
        dismiss_text = trans("dialog.rename.dismiss")
        edit_title = trans('action.edit')

        for dict_id, data in self.dicts.items():
            parent_id = f"{self.id}.{dict_id}"
            option_key = self.keys[dict_id]
            parent = self.parents[dict_id]
            ui.config[parent_id] = {}

            fields = {}
            if data["type"] == 'dict':
                fields = controller.config.dictionary.to_options(
                    parent_id,
                    data,
                )
            elif data["type"] == 'cmd':
                fields = controller.config.cmd.to_options(
                    parent_id,
                    data,
                )

            widgets = self.build_widgets(
                parent_id,
                fields,
                stretch=True,
            )
            ui.config[parent_id] = widgets

            add_option = self.add_option
            add_row_option = self.add_row_option
            add_raw_option = self.add_raw_option

            rows = QVBoxLayout()
            has_textarea = False

            for k, f in fields.items():
                t = f.get("type")
                w = widgets.get(k)
                if t in ('int', 'float'):
                    row = add_option(w, f)
                elif t in ('text', 'textarea', 'dict', 'combo'):
                    row = add_row_option(w, f)
                elif t == 'bool':
                    row = add_raw_option(w, f)
                else:
                    continue
                rows.addLayout(row)
                if t == 'textarea':
                    has_textarea = True

            if not has_textarea:
                rows.addStretch()

            save_key = f"{parent_id}.btn.save"
            dismiss_key = f"{parent_id}.btn.dismiss"

            ui.nodes[save_key] = QPushButton(save_text)
            ui.nodes[save_key].clicked.connect(
                lambda checked=True, option_key=option_key, parent=parent, fields=fields:
                controller.config.dictionary.save_editor(
                    option_key,
                    parent,
                    fields,
                ))
            ui.nodes[save_key].setAutoDefault(True)

            ui.nodes[dismiss_key] = QPushButton(dismiss_text)
            ui.nodes[dismiss_key].clicked.connect(
                lambda checked=True, pid=parent_id: ui.dialogs.close(f'editor.{pid}')
            )
            ui.nodes[dismiss_key].setAutoDefault(False)

            footer = QHBoxLayout()
            footer.addWidget(ui.nodes[dismiss_key])
            footer.addWidget(ui.nodes[save_key])

            scroll = QScrollArea()
            scroll.setWidgetResizable(True)
            scroll.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

            widget = QWidget()
            widget.setLayout(rows)
            scroll.setWidget(widget)
            widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

            layout = QVBoxLayout()
            layout.addWidget(scroll)
            layout.addLayout(footer)

            dialog_key = f'editor.{parent_id}'
            ui.dialog[dialog_key] = EditorDialog(self.window, parent_id)
            ui.dialog[dialog_key].setLayout(layout)
            ui.dialog[dialog_key].setWindowTitle(edit_title)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.05 18:00:00                  #
# ================================================== #

import json
from dataclasses import dataclass
from typing import Optional


@dataclass(slots=True)
class PromptItem:
    id: Optional[object] = None
    name: Optional[str] = None
    content: Optional[str] = None

    def __init__(self):
        """
        Prompt item
        """
        self.id = None
        self.name = None
        self.content = None

    def serialize(self) -> dict:
        """
        Serialize item to dict

        :return: serialized item
        """
        return {
            'id': str(self.id),
            'name': self.name,
            'content': self.content,
        }

    def deserialize(self, data: dict):
        """
        Deserialize item from dict

        :param data: serialized item
        """
        if 'id' in data:
            self.id = data['id']
        if 'name' in data:
            self.name = data['name']
        if 'content' in data:
            self.content = data['content']

    def dump(self) -> str:
        """
        Dump item to string

        :return: serialized item
        """
        try:
            return json.dumps(self.serialize())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        """To string"""
        return self.dump()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.05 20:00:00                  #
# ================================================== #

from typing import Any, Optional


def as_int(val: Any) -> Optional[int]:
    """
    Coerce to int if possible, else None.

    :param val: Input value
    :return: int or None
    """
    if val is None:
        return None
    try:
        return int(val)
    except Exception:
        try:
            return int(float(val))
        except Exception:
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.28 04:00:00                  #
# ================================================== #

import copy
import os
from pathlib import Path
from typing import Optional, Union
from uuid import uuid4

from PySide6.QtCore import QObject, QRunnable, Signal, Slot
from PySide6.QtWidgets import QApplication

from pygpt_net.utils import trans


class WorkerSignals(QObject):
    """
    Signals for worker to communicate with main thread.
    """
    finished = Signal()  # when worker is finished
    deleted = Signal(str)  # after deleting profile
    duplicated = Signal(str, str)  # after duplicating profile (str: new_uuid, str: new_name)
    error = Signal(str)  # when error occurs
    hideStatus = Signal()  # hide status in dialog
    updateStatus = Signal(str)    # update status in dialog
    updateGlobalStatus = Signal(str)  # update global status
    alert = Signal(object)        # dialog alert
    after_migrate = Signal(bool, str, str, str)  # result, profile_name, current_path, new_path
    confirm = Signal(str, str, str)  # confirm dialog (action, path, message)
    restored = Signal(str)        # after restoring workdir
    updated = Signal(str)         # after updating workdir
    switch = Signal(str)       # switch to profile (after reset)


class WorkdirWorker(QRunnable):
    """Worker for handling workdir operations in a separate thread."""
    def __init__(
            self,
            window,
            action: str,
            path: str = None,
            force: bool = False,
            current: Optional[str] = None,
            profile_name: Optional[str] = None,
            profile_uuid: Optional[Union[str, list]] = None,
            profile_new_name: Optional[str] = None,
            profile_new_path: Optional[str] = None,
            batch: bool = False,
    ):
        super().__init__()
        self.window = window
        self.action = action
        self.path = path
        self.force = force
        self.current = current
        self.profile_name = profile_name
        self.profile_uuid = profile_uuid
        self.profile_new_name = profile_new_name
        self.profile_new_path = profile_new_path
        self.signals = WorkerSignals()
        self.batch = batch

    @Slot()
    def run(self):
        try:
            if self.action == "migrate":
                self.worker_migrate()
            elif self.action == "delete":
                self.worker_delete_files()
            elif self.action == "duplicate":
                self.worker_duplicate()
            elif self.action == "reset":
                self.worker_reset()
            else:
                self.signals.error.emit(f"Unknown action: {self.action}")

        except Exception as e:
            self.window.core.debug.log(e)
            self.signals.error.emit(str(e))

        finally:
            self.signals.finished.emit()
            self.cleanup()

    def cleanup(self):
        """Cleanup resources after worker execution."""
        sig = self.signals
        self.signals = None
        if sig is not None:
            try:
                sig.deleteLater()
            except RuntimeError:
                pass

    def worker_delete_files(self):
        """Delete files and directories associated with the profile"""
        uuid = self.profile_uuid
        profiles = self.window.controller.settings.profile.get_profiles()
        remove_datadir = True
        remove_db = True
        uuids = uuid if isinstance(uuid, list) else [uuid]
        for uuid in uuids:
            if uuid in profiles:
                profile = profiles[uuid]
                name = profile['name']
                path = profile['workdir'].replace("%HOME%", str(Path.home()))
                # remove profile
                if self.window.core.config.profile.remove(uuid):
                    if not os.path.exists(path) or not os.path.isdir(path):
                        if not self.batch:
                            self.signals.deleted.emit(name)
                            self.signals.alert.emit(trans("dialog.profile.alert.path.not_exists"))
                            return
                        else:
                            print("Directory not exists, ignoring: ", path)
                            self.signals.updateGlobalStatus.emit(f"Directory not exists: {path}")
                            self.signals.deleted.emit(name)
                            return # nothing to delete
                    print(f"Clearing workdir: {path}")
                    try:
                        self.window.core.filesystem.clear_workdir(
                            path,
                            remove_db=remove_db,
                            remove_datadir=remove_datadir,
                        )
                    except Exception as e:
                        print("Error deleting profile files: ", e)
                    self.signals.deleted.emit(name)

    def worker_duplicate(self):
        """Duplicate profile"""
        uuid = self.profile_uuid
        new_name = self.profile_new_name
        new_path = self.profile_new_path

        profiles = self.window.controller.settings.profile.get_profiles()
        if uuid not in profiles:
            self.signals.alert.emit("Profile not found!")
            return
        profile = profiles[uuid]

        copy_datadir = self.window.controller.settings.profile.is_include_datadir()
        copy_db = self.window.controller.settings.profile.is_include_db()

        # make copy
        duplicate = copy.deepcopy(profile)
        new_uuid = str(uuid4())
        duplicate['name'] = new_name
        duplicate['workdir'] = new_path
        self.window.core.config.profile.append(new_uuid, duplicate)

        # copy files from workdir
        path_from = profile['workdir'].replace("%HOME%", str(Path.home()))
        path_to = new_path
        print(f"Copying all files from {path_from} to: {path_to}")
        self.signals.updateGlobalStatus.emit("Copying files...")
        result = self.window.core.filesystem.copy_workdir(
            path_from,
            path_to,
            copy_db=copy_db,
            copy_datadir=copy_datadir,
        )
        if not result:
            self.signals.alert.emit("Error copying files!")
            self.signals.updateGlobalStatus.emit("Error copying files!")
            return
        print("[OK] All files copied successfully.")
        self.signals.duplicated.emit(new_uuid, new_name)

    def worker_reset(self):
        """Reset profile"""
        uuid = self.profile_uuid
        profiles = self.window.controller.settings.profile.get_profiles()
        current = self.window.core.config.profile.get_current()
        remove_datadir = False
        remove_db = True
        uuids = uuid if isinstance(uuid, list) else [uuid]
        for uuid in uuids:
            if uuid in profiles:
                profile = profiles[uuid]
                path = profile['workdir'].replace("%HOME%", str(Path.home()))
                if not os.path.exists(path) or not os.path.isdir(path):
                    if not self.batch:
                        self.signals.alert.emit(f"Directory not exists: {path}")
                        return
                    else:
                        print("Directory not exists, ignoring: ", path)
                        self.signals.updateGlobalStatus.emit(f"Directory not exists: {path}")
                        return
                print("Clearing workdir: ", path)
                if uuid == current:
                    self.window.core.db.close()
                self.window.core.filesystem.clear_workdir(
                    path,
                    remove_db=remove_db,
                    remove_datadir=remove_datadir,
                )
                if uuid == current and not self.batch:
                    self.signals.switch.emit(uuid)  # switch to profile
                self.signals.updateGlobalStatus.emit(f"Profile cleared: {profile['name']}")

    def worker_migrate(self):
        """Migrate working directory"""
        if self.window.controller.settings.workdir.busy:
            self.signals.alert.emit("Workdir migration in progress...")
            return

        # check if not the same directory
        current = self.window.core.config.get_user_path()
        if current == self.path:
            self.signals.alert.emit(trans("dialog.workdir.result.same_directory"))
            self.window.controller.settings.workdir.busy = False
            self.signals.hideStatus.emit()
            return

        # check if path is not already a workdir
        if self.window.core.filesystem.is_workdir_in_path(self.path):
            msg = trans("dialog.workdir.update.confirm").format(path=self.path)
            self.signals.confirm.emit("workdir.update", self.path, msg)
            return

        # check if path is empty
        if not self.window.core.filesystem.is_directory_empty(self.path):
            self.signals.alert.emit(trans("dialog.workdir.change.empty.alert"))
            return

        # confirm move
        if not self.force:
            msg = trans("dialog.workdir.change.confirm").format(path=self.path)
            self.signals.confirm.emit("workdir.change", self.path, msg)
            return

        self.signals.hideStatus.emit()
        self.window.controller.settings.workdir.busy = True
        print(f"Migrating workdir from: {current} to: {self.path}...")

        # check if path exists
        if not os.path.exists(self.path) or not os.path.isdir(self.path):
            self.signals.alert.emit(trans("dialog.workdir.result.directory_not_exists"))
            self.signals.hideStatus.emit()
            self.window.controller.settings.workdir.busy = False
            return

        # check free space
        space_required = self.window.core.filesystem.get_directory_size(current, human_readable=False)
        space_free = self.window.core.filesystem.get_free_disk_space(self.path, human_readable=False)
        if space_required > space_free:
            self.signals.alert.emit(trans("dialog.workdir.result.no_free_space").format(
                required=self.window.core.filesystem.get_directory_size(current),
                free=self.window.core.filesystem.get_free_disk_space(self.path),
            ))
            self.window.controller.settings.workdir.busy = False
            self.signals.hideStatus.emit()
            return

        # copy workdir
        self.signals.updateGlobalStatus.emit(trans("dialog.workdir.result.wait"))
        QApplication.processEvents()  # process events to update UI
        try:
            result = self.window.core.filesystem.copy_workdir(current, self.path)
        except Exception as e:
            self.window.core.debug.log(e)
            self.signals.alert.emit(str(e))
            print("Error migrating workdir: ", e)
            result = False

        # reload UI, config, etc.
        self.signals.after_migrate.emit(result, self.profile_name, current, self.path)


class Workdir:
    def __init__(self, window=None):
        """
        Workdir controller

        :param window: window instance
        """
        self.window = window
        self.is_dialog = False
        self.busy = False

    def rollback(self, current: str = None):
        """
        Rollback to the previous working directory

        :param current: current working directory (optional)
        """
        print(f"Reverting workdir to: {current}...")
        self.window.core.config.set_workdir(current, reload=True)
        default_path = self.window.core.config.get_base_workdir()
        lock_file = os.path.join(default_path, 'path.cfg')
        lock_path = current.replace(str(Path.home()), "%HOME%")
        if current == default_path:
            lock_path = ""
        with open(lock_file, 'w', encoding='utf-8') as f:
            f.write(lock_path)
        self.window.ui.dialogs.workdir.set_path(current)
        self.window.ui.dialogs.workdir.show_status(f"Failed. Reverted to previous workdir: {current}.")
        self.window.controller.reload()
        self.window.core.config.profile.update_current_workdir(current)

    def update_workdir(
            self,
            force: bool = False,
            path: str = None,
            is_create: bool = False,
    ):
        """
        Update working directory

        :param force: boolean indicating if update should be forced (confirm)
        :param path: new working directory to set
        :param is_create: True if called on profile creation
        """
        print("\n====================")
        print(f"Changing workdir to: {path}")
        print("====================\n")
        default_path = self.window.core.config.get_base_workdir()
        if force:
            self.window.ui.dialogs.workdir.show_status(trans("dialog.workdir.result.wait"))

        lock_file = os.path.join(default_path, 'path.cfg')  # put "path.cfg"
        lock_path = path.replace(str(Path.home()), "%HOME%")
        if path == default_path:
            lock_path = ""  # set empty if default dir
        with open(lock_file, 'w', encoding='utf-8') as f:
            f.write(lock_path)

        # update path in current profile
        self.window.core.config.profile.update_current_workdir(path)

        # save previous theme and language to retain them after workdir change
        prev_theme = None
        prev_lang = None
        if is_create:
            prev_theme = self.window.core.config.get('theme')
            prev_lang = self.window.core.config.get('lang')

        # reload config
        self.window.core.config.set_workdir(path, reload=True)

        # if profile is just created, use current theme and language
        if is_create:
            print("Using current theme and language: ", prev_theme, prev_lang)
            if prev_theme is not None:
                self.window.core.config.set('theme', prev_theme)
            if prev_lang is not None:
                self.window.core.config.set('lang', prev_lang)
            self.window.core.config.save()

        self.window.core.config.set('license.accepted', True)  # accept license to prevent show dialog again

    @Slot(bool, str, str, str)
    def do_update(
            self,
            force: bool,
            profile_name: str,
            current_path: str,
            new_path: str,
            is_create: bool = False
    ) -> bool:
        """
        Update working directory

        :param force: boolean indicating if update should be forced (confirm)
        :param profile_name: profile name to update after workdir change
        :param current_path: current working directory before update
        :param new_path: new working directory to set
        :param is_create: if True, skip check for existing workdir in path
        :return: boolean indicating if update was successful
        """
        self.update_workdir(
            force=force,
            path=new_path,
            is_create=is_create,
        )
        rollback = False
        success = False
        if force:
            try:
                self.window.ui.dialogs.workdir.show_status(trans("dialog.workdir.result.wait"))
                self.window.controller.reload()  # reload all
                self.window.ui.dialogs.workdir.show_status(trans("dialog.workdir.result.wait"))
                msg = trans("dialog.workdir.result.success").format(path=new_path)
                self.window.ui.dialogs.workdir.show_status(msg)
                self.window.ui.dialogs.alert(msg)
                success = True
            except Exception as e:
                rollback = True
                self.window.core.debug.log(e)
                self.window.ui.dialogs.alert(str(e))
                print("Error reloading components: ", e)
                self.window.controller.reloading = False  # unlock
        else:
            self.window.controller.reload()  # reload only

        if rollback:  # if failed
            self.rollback(current=current_path)  # revert to previous workdir
        else:
            # update profile after workdir change
            if profile_name:
                self.window.controller.settings.profile.after_update(profile_name)
        return success

    @Slot(bool, str, str, str)
    def do_migrate(
            self,
            result: bool,
            profile_name: str,
            current_path: str,
            new_path: str
    ) -> bool:
        """
        Handle migration result

        :param result: boolean indicating if migration was successful
        :param profile_name: profile name to update after migration
        :param current_path: current working directory before migration
        :param new_path: new working directory after migration
        :return: boolean indicating if migration was successful
        """
        success = False
        if result:
            try:
                # update workdir to new path
                success = self.do_update(
                    force=True,
                    profile_name=profile_name,
                    current_path=current_path,
                    new_path=new_path,
                )  # with rollback if failed

                if not success:
                    raise Exception("Migration failed, workdir not updated.")

                msg = trans("dialog.workdir.result.success").format(path=new_path)
                self.window.ui.dialogs.workdir.show_status(msg)
                self.window.ui.dialogs.alert(msg)

                # remove old workdir only if success
                self.window.core.debug.info(f"Clearing old workdir: {current_path}...")
                try:
                    self.window.core.filesystem.clear_workdir(current_path)  # allow errors here
                    self.window.core.debug.info(f"Old workdir cleared: {current_path}.")
                except Exception as e:
                    self.window.core.debug.log(e)
                    print("Error clearing old workdir: ", e)

            except Exception as e:
                self.window.core.debug.log(e)
                self.window.ui.dialogs.alert(str(e))
                print("Error migrating workdir: ", e)
                self.window.controller.reloading = False
        else:
            # if migration failed
            self.window.ui.dialogs.workdir.show_status(trans("dialog.workdir.result.failed"))
            self.window.ui.dialogs.alert(trans("dialog.workdir.result.failed"))
            self.window.controller.reloading = False

        self.window.controller.settings.workdir.busy = False
        if success:
            self.window.core.debug.info(f"Finished migrating workdir from: {current_path} to: {new_path}.")
        return success

    def change(self):
        """Change working directory (open dialog)"""
        if self.is_dialog:
            self.window.ui.dialogs.close('workdir.change')
            self.is_dialog = False
        else:
            self.window.ui.dialogs.workdir.prepare()
            self.window.ui.dialogs.workdir.set_path(self.window.core.config.get_user_path())
            self.window.ui.dialogs.workdir.hide_status()
            self.window.ui.dialogs.open(
                'workdir.change',
                width=600,
                height=180,
            )
            self.is_dialog = True

    def run_action(
            self,
            action: str,
            path: str = None,
            force: bool = False,
            current: str = None,
            profile_name: Optional[str] = None,
            profile_uuid: Optional[Union[str, list]] = None,
            profile_new_name: Optional[str] = None,
            profile_new_path: Optional[str] = None,
            batch: bool = False,
    ):
        """
        Run action in thread

        :param action: action to perform (update, migrate, restore)
        :param path: path for update or migrate
        :param force: force action (confirm)
        :param current: current working directory (for restore action)
        :param profile_name: profile name (optional, for after update callback)
        :param profile_uuid: profile UUID (optional, for delete files action) or list of UUIDs
        :param profile_new_name: new profile name (optional, for duplicate action)
        :param profile_new_path: new profile path (optional, for duplicate action)
        :param batch: if True, run in batch mode (no UI updates)
        """
        worker = WorkdirWorker(
            window=self.window,
            action=action,
            path=path,
            force=force,
            current=current,
            profile_name=profile_name,
            profile_uuid=profile_uuid,
            profile_new_name=profile_new_name,
            profile_new_path=profile_new_path,
            batch=batch,
        )

        # connect signals
        worker.signals.updateGlobalStatus.connect(self.window.update_status)
        worker.signals.updateStatus.connect(self.window.ui.dialogs.workdir.show_status)
        worker.signals.hideStatus.connect(self.window.ui.dialogs.workdir.hide_status)
        worker.signals.alert.connect(self.window.ui.dialogs.alert)
        worker.signals.error.connect(lambda err: self.window.core.debug.log(f"Worker error: {err}"))
        worker.signals.confirm.connect(self.window.ui.dialogs.confirm)
        worker.signals.restored.connect(lambda current: self.window.ui.dialogs.workdir.set_path(current))
        worker.signals.updated.connect(self.window.controller.settings.profile.after_update)
        worker.signals.deleted.connect(self.window.controller.settings.profile.after_delete)
        worker.signals.duplicated.connect(self.window.controller.settings.profile.after_duplicate)
        worker.signals.switch.connect(self.window.controller.settings.profile.switch_current)
        worker.signals.after_migrate.connect(self.do_migrate)

        # start worker in thread pool
        self.window.threadpool.start(worker)

    def update(
            self,
            path: str,
            force: bool = False,
            profile_name: str = None,
            is_create: bool = False,
    ):
        """
        Switch working directory to the existing one

        :param path: existing working directory
        :param force: force update (confirm)
        :param profile_name: profile name (optional, for future use)
        :param is_create: if True, skip check for existing workdir in path
        """
        self.do_update(
            force=force,
            profile_name=profile_name,
            current_path=self.window.core.config.get_user_path(),
            new_path=path,
            is_create=is_create,
        )

    def migrate(
            self,
            path: str,
            force: bool = False
    ):
        """
        Migrate working directory

        :param path: new working directory
        :param force: force migration (confirm)
        """
        self.run_action(
            action="migrate",
            path=path,
            force=force,
        )

    def restore(
            self,
            current: str
    ):
        """
        Restore default working directory

        :param current: current working directory
        """
        self.rollback(current=current)

    def delete_files(
            self,
            profile_uuid: Union[str, list],
            batch: bool = False
    ):
        """
        Delete files and directories associated with the profile

        :param profile_uuid: profile UUID
        :param batch: if True, run in batch mode (no UI updates)
        """
        self.run_action(
            action="delete",
            profile_uuid=profile_uuid,
            batch=batch,
        )

    def duplicate(
            self,
            profile_uuid: str,
            new_name: str,
            new_path: str
    ):
        """
        Duplicate profile

        :param profile_uuid: profile UUID to duplicate
        :param new_name: new profile name
        :param new_path: new profile path
        """
        self.run_action(
            action="duplicate",
            profile_uuid=profile_uuid,
            profile_new_name=new_name,
            profile_new_path=new_path,
        )

    def reset(
            self,
            profile_uuid: Union[str, list],
            batch: bool = False
    ):
        """
        Reset profile

        :param profile_uuid: profile UUID to reset
        :param batch: if True, run in batch mode (no UI updates)
        """
        self.run_action(
            action="reset",
            profile_uuid=profile_uuid,
            batch=batch,
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.05 17:00:00                  #
# ================================================== #

import datetime
from typing import Optional, List, Dict, Any

from packaging.version import Version

from pygpt_net.item.store import RemoteStoreItem
from pygpt_net.provider.core.remote_store.db_sqlite import DbSqliteProvider

from .files import Files


class Store:
    """
    Anthropic Files API "store" adapter.
    Anthropic does not have vector stores; we expose a single workspace-level pseudo store.
    """

    PROVIDER_NAME = "anthropic"
    DEFAULT_STORE_ID = "files"
    DEFAULT_STORE_NAME = "Files"

    def __init__(self, window=None):
        """
        Anthropic Files workspace core

        :param window: Window instance
        """
        self.window = window
        self.provider = DbSqliteProvider(window)
        self.files = Files(window)
        self.items: Dict[str, RemoteStoreItem] = {}

    def install(self):
        """Install provider data"""
        self.provider.install()
        self.files.install()

    def patch(self, app_version: Version) -> bool:
        """Patch provider data"""
        res1 = self.files.patch(app_version)
        res2 = self.provider.patch(app_version)
        return res1 or res2

    def get(self, id: str) -> RemoteStoreItem:
        if id in self.items:
            return self.items[id]

    def get_ids(self) -> List[str]:
        return list(self.items.keys())

    def get_all(self) -> Dict[str, RemoteStoreItem]:
        return self.items

    def get_id_by_idx_all(self, idx: int) -> str:
        return list(self.items.keys())[idx]

    def get_by_idx(self, idx: int) -> str:
        items = self.items
        return list(items.keys())[idx]

    def has(self, id: str) -> bool:
        return id in self.items

    def _ensure_default(self) -> RemoteStoreItem:
        """Ensure default pseudo-store exists locally."""
        if self.DEFAULT_STORE_ID in self.items:
            return self.items[self.DEFAULT_STORE_ID]
        store = RemoteStoreItem()
        store.id = self.DEFAULT_STORE_ID
        store.name = self.DEFAULT_STORE_NAME
        store.provider = self.PROVIDER_NAME
        store.is_thread = False
        store.record_id = self.provider.create(store)
        self.items[store.id] = store
        return store

    def create(self, name: Optional[str] = None) -> Optional[RemoteStoreItem]:
        """
        Create workspace-level pseudo-store (idempotent).

        :param name: local alias
        :return: store item
        """
        store = self._ensure_default()
        if name:
            store.name = name
            self.provider.save(store)
        return store

    def update(self, store: RemoteStoreItem) -> Optional[RemoteStoreItem]:
        """
        Update local store metadata (no remote update).
        """
        self.items[store.id] = store
        self.provider.save(store)
        return store

    def get_status_data(self, id: str):
        """
        Compute workspace status from Files API
        """
        status = {}
        stats = self.window.core.api.anthropic.store.get_files_stats()
        if stats is not None:
            status = {
                "status": "ready",
                "usage_bytes": int(stats.get("total_bytes", 0) or 0),
                "expires_at": None,
                "last_active_at": int(datetime.datetime.now().timestamp()),
                "file_counts": {
                    "in_progress": 0,
                    "completed": int(stats.get("count", 0) or 0),
                    "cancelled": 0,
                    "failed": 0,
                    "total": int(stats.get("count", 0) or 0),
                },
                "expires_after": None,
                "remote_display_name": self.items.get(id, RemoteStoreItem()).name if id in self.items else self.DEFAULT_STORE_NAME,
            }
        return status, None

    def update_status(self, id: str):
        """
        Update pseudo-store status
        """
        self._ensure_default()
        store = self.items[id]
        status, _ = self.get_status_data(id)
        self.append_status(store, status)
        self.update(store)

    def append_status(self, store: RemoteStoreItem, status: Dict[str, Any]):
        now = datetime.datetime.now()
        ts = int(now.timestamp())
        status["__last_refresh__"] = now.strftime("%Y-%m-%d %H:%M:%S")
        store.status = status
        store.last_sync = ts
        if "status" in status:
            store.last_status = status["status"]
        if "usage_bytes" in status:
            store.usage_bytes = status["usage_bytes"]
        if "file_counts" in status:
            store.num_files = status["file_counts"]["total"]
        if "last_active_at" in status and status["last_active_at"]:
            store.last_active = int(status["last_active_at"])

    def get_names(self) -> Dict[str, str]:
        names = {}
        for id in self.items:
            store = self.items[id]
            names[id] = store.name
        return names

    def delete(self, id: str) -> bool:
        if id in self.items:
            store = self.items[id]
            self.provider.delete_by_id(store.record_id)
            del self.items[id]
            return True
        return False

    def import_items(self, items: Dict[str, RemoteStoreItem]):
        self.items = items
        for item in items.values():
            item.provider = self.PROVIDER_NAME
            item.record_id = self.provider.create(item)

    def clear(self):
        self.truncate()

    def is_hidden(self, id: str) -> bool:
        return False

    def truncate(self) -> bool:
        self.provider.truncate(self.PROVIDER_NAME)
        self.items = {}
        return True

    def load(self):
        self.items = self.provider.load_all(self.PROVIDER_NAME)
        if self.DEFAULT_STORE_ID not in self.items:
            # Ensure default exists
            self._ensure_default()
        self.sort_items()

    def load_all(self):
        self.load()
        self.files.load()

    def sort_items(self):
        pass

    def save(self):
        self.provider.save_all(self.items)

    def get_version(self) -> str:
        return self.provider.get_version()